"""Re-usable state segments."""

from contextlib import contextmanager
from functools import partial, update_wrapper
from inspect import BoundArguments, Signature, signature
from math import ceil, log2
from typing import (
    Any,
    Callable,
    ClassVar,
    Concatenate,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Protocol,
    Self,
    Sequence,
    Tuple,
    Union,
    overload,
)

from nortl.core.constructs.utils import FastForwarder
from nortl.core.modifiers import BaseModifier, WeakReference
from nortl.core.operations import Const, Var, to_renderable
from nortl.core.protocols import (
    AnySignal,
    EngineProto,
    MemoryViewProto,
    MemoryZoneProto,
    PermanentSignal,
    Renderable,
    ScratchSignalProto,
    SignalProto,
    StateProto,
    WorkerProto,
)
from nortl.core.signal import ScratchSignal, Signal, SignalSlice

__all__ = [
    'Segment',
]


class HelperProto(Protocol):
    @property
    def engine(self) -> EngineProto: ...


class Ignore:
    """Marker for ignored call arguments."""

    _inst: ClassVar['Ignore']

    def __new__(cls) -> 'Ignore':
        if not hasattr(cls, '_inst'):
            cls._inst = object.__new__(cls)
        return cls._inst

    def __repr__(self) -> str:
        return '...'


class RenderedSegment:
    """Helper class to store information about a rendered Segment for a specific call signature and worker."""

    def __init__(
        self,
        start_state: StateProto,
        end_state: StateProto,
        input_slots: Mapping[str, ScratchSignalProto],
        output_slot_widths: Sequence[Optional[int]],
        result: object,
        memory_view: MemoryViewProto,
    ) -> None:
        self.start_state = start_state
        self.end_state = end_state
        self.input_slots = input_slots
        self.output_slot_widths = output_slot_widths
        self.result: object = result
        self.memory_view = memory_view
        self.calls: List[Tuple[StateProto, StateProto, Sequence[WeakReference[SignalProto]]]] = []


class EngineContext:
    """Helper class to store information about all segments for a specific engine."""

    def __init__(self, engine: EngineProto) -> None:
        self.engine = engine
        self.return_addresses: Dict[WorkerProto, SignalProto] = {}
        self._segments: List[SegmentImplementation[Any, Any, Any]] = []

    @property
    def current_worker(self) -> WorkerProto:
        """Selected worker."""
        return self.engine.current_worker

    @property
    def return_address(self) -> SignalProto:
        """Return address for current worker."""
        return self.return_addresses[self.current_worker]

    def get_return_address(self, min_calls: int) -> SignalProto:
        """Get signal that stores the return address for the current worker.

        Expands the signal if needed, to fit the minimum number of calls.
        """
        min_width = max(1, ceil(log2(min_calls)))

        if self.current_worker not in self.return_addresses:
            # Create new return address
            name = self.current_worker.create_scoped_name('RETURN_ADDRESS')
            return_address = self.engine.define_local(name, width=Var(min_width), reset_value=0)

            self.return_addresses[self.current_worker] = return_address
        else:
            # Get signal that stores the return address
            return_address = self.return_addresses[self.current_worker]

            # Resize it to support the number of calls for this level, if necessary
            width: Var = return_address.width  # type: ignore[assignment]
            if width.value < min_width:
                width.update(min_width)
        return return_address

    def register_segment[T: Union[EngineProto, HelperProto], **P, R: Optional[Union[AnySignal, Sequence[AnySignal]]]](
        self, segment: 'SegmentImplementation[T, P, R]'
    ) -> None:
        """Register a segment, if it isn't registered yet. All segments belonging to a worker are tracked for debugging."""
        if segment not in self._segments:
            self._segments.append(segment)

    @property
    def segments(self) -> Sequence['SegmentImplementation[Any, Any, Any]']:
        """Sequence of all segments."""
        return self._segments


SlotSize = Union[int, str, bool]


class SlotSpec:
    """Specification for input and output slots."""

    def __init__(self, width: SlotSize) -> None:
        self.width = width

    def __format__(self, format_spec: str) -> str:
        if isinstance(self.width, bool):
            if self.width:
                return 'Automatic detection based on the operand width of the first argument passed into the slot.'
            else:
                return 'Disabled.'
        elif isinstance(self.width, int):
            return f'Fixed width of {self.width} Bit.'
        else:
            return f'Parametric width based on the value of argument {self.width}'

    def __repr__(self) -> str:
        return f'<SlotSpec (width={self.width})>'


class InputCheck:
    """Input check callback."""

    def __init__(self, callback: Callable[[BoundArguments], bool], description: str):
        self.callback = callback
        self.description = description

    def __call__(self, bound_arguments: BoundArguments) -> bool:
        return self.callback(bound_arguments)

    @classmethod
    def from_validator[**P](cls, validator: Callable[P, bool]) -> Self:
        """Create input check from validator callback."""

        def callback(bound_arguments: BoundArguments) -> bool:
            return validator(*bound_arguments.args[1:], **bound_arguments.kwargs)

        return cls(callback, f'Validator callback {validator.__name__}({", ".join(signature(validator).parameters.keys())})')

    @classmethod
    def for_argument[T](cls, name: str, check: Callable[[T], bool]) -> Self:
        """Create input check for a single argument."""

        def callback(bound_arguments: BoundArguments) -> bool:
            value = bound_arguments.arguments[name]
            return check(value)

        return cls(callback, f'Validation for {name}')


class semistaticmethod[T, **P, R]:  # noqa: N801
    """Turns method into a semi-static method.

    It will allow calls both as a instance method, but also as a static method of the class. In the later case, the self-argument will be None.
    """

    def __init__(self, method: Callable[Concatenate[T, P], R]):
        self.__method = method

    def __get__(self, obj: Optional[T], objtype: Optional[type] = None) -> Union[Self, Callable[P, R]]:
        # Non-data descriptor, this allows the decorator to be used on class methods
        if obj is None:
            return self
        else:
            return partial(self.__method, obj)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        return partial(self.__method, None)(*args, **kwargs)  # type: ignore[arg-type]


class SegmentFactoryMixin:
    """Helper class to create segments with non-default configuration.

    This is used by both the Segment class itself and the SegmentFactory.
    """

    @semistaticmethod
    def with_input_slots(self, **input_slots: SlotSize) -> 'SegmentFactory':
        """Creates a segment with input slots for the selected inputs.

        Input slots are scratch signals, to which the input arguments are copied, before calling the segment.
        They help, if the same segment is meant to be re-used for many different inputs, without explicitely defining a global signal for the input.

        Arguments:
            input_slots: Keywords mapping the argument names of the decorated function or method to slot sizes.

        Slot sizes can be specified with the following options:

        - If the slot is an integer value, it specifies the width of the input slot in bits.
        - If the slot size is a string, it refers to another argument of the decorated function or method.
            The slot size is set to the value of this argument, when called. This allows defining parametric segments.
        - If the slot size is `True`, the slot size will be automatically determined, when it the segment is called for the first time.
            The input value needs to have an fixed `operand_width`, for this to work.
        - If the slot size is `False`, no slot is added. This is redundant, as no slot is added by default.

        It is recommended to avoid using a slot size of `True`, as the automatic detection will fail, if passing in any dynamic data.

        Example:
            The following example defines an segment with an input argument `data`. A slot with a fixed size of 8 bit is added.

            ```python
            from nortl import Engine, Segment, to_renderable

            # Define signals
            engine = Engine('my_engine')
            output = engine.define_output('output', width=1)

            @Segment.with_input_slots(data=8)
            def send_byte(engine: EngineProto, data: Union[Renderable, int]) -> None:
                data = to_renderable(data)

                for i in range(8):
                    engine.set(output, data[i])
                    engine.sync()
            ```

            Note how the decorated function deserializes the data in a for-loop. The length is currently fixed to 8.

            If the segment should accept dynamic lengths of data, a `width` argument can be added.
            Now, the width of the input slot for `data` needs to match it.
            This is achieved by setting the slot size to `'width'`:

            ```python
            # ...

            @Segment.with_input_slots(data='width')
            def send_dynamic(engine: EngineProto, data: Union[Renderable, int], width: int) -> None:
                data = to_renderable(data)

                for i in range(width):
                    engine.set(output, data[i])
                    engine.sync()
            ```

        Input slots are located inside the memory zone of a segment.
        """
        return SegmentFactory(
            input_checks=getattr(self, '_SegmentFactory__input_checks', (None, {})),
            input_slots=input_slots,
            output_slots=getattr(self, '_SegmentFactory__output_slots', ()),
        )

    @semistaticmethod
    def with_output_slots(self, *output_slots: SlotSize) -> 'SegmentFactory':
        """Creates a segment with output slots for the selected outputs.

        Output slots are scratch signals, to which the result(s) are copied.
        By default, any signals returned from segments point to the same register and are therefore read-only.
        If an output slot is added, the result will be copied into a scratch signal for each call.

        Arguments:
            output_slots: Slot sizes for the output values, in the same order as they are returned.

        Slot sizes can be specified with the following options:

        - If the slot is an integer value, it specifies the width of the output slot in bits.
        - If the slot size is a string, it refers to an argument of the decorated function or method.
            The slot size is set to the value of this argument, when called. This allows defining parametric segments.
        - If the slot size is `True`, the slot size will be automatically determined, when it the segment is called for the first time.
            The result needs to have an fixed `operand_width`, for this to work.
        - If the slot size is `False`, no slot is added. This is redundant, as no slot is added by default.

        It is recommended to avoid using a slot size of `True`, as the automatic detection will fail, if passing in any dynamic data.

        Output slots are located inside the memory zone of the caller.
        """
        return SegmentFactory(
            input_checks=getattr(self, '_SegmentFactory__input_checks', (None, {})),
            input_slots=getattr(self, '_SegmentFactory__input_slots', {}),
            output_slots=output_slots,
        )

    @semistaticmethod
    def with_input_checks[**P](self, input_validator: Optional[Callable[P, bool]] = None, /, **input_checks: Callable[..., bool]) -> 'SegmentFactory':
        """Creates a segment with input checks.

        Input checks are callback functions the validate the arguments passed to a segment.

        If a segment uses input slots, the arguments with input slots can no longer be validated inside the function or methods body, because it will not be called
        for different inputs. Input checks solve the issue.

        Arguments:
            input_validator: A validator callback, that accepts all arguments and keyword arguments of the decorated function or method.
                The first argument (engine or self) must be omitted.
                The validator callback may either return a boolean result, where True means that all arguments passed validation.
                Or it may directly raise an exception.
                As the validator callback receives all arguments at the same time, it can be used for "global" checks.
            input_checks: Keywords mapping the argument names of the decorated function or method to individual validators, e.g. created by lambda
                functions.
                The input checks receives only one argument at a time, so they cannot be used for "global" checks.
                However, they can be written "in-line", by using lambda functions.


        Examples:
            The following segment can send dynamic amounts of data, without using a ForLoop construct. However, only up to 31 bits are supported at a time.
            To validate that the length never exceeds 31, a validator callback is added.

            ```python
            from nortl import Engine, Segment, to_renderable

            # Define signals
            engine = Engine('my_engine')
            output = engine.define_output('output', width=1)

            def validate_length(data: int, length: int = 16) -> bool:
                return length < 32

            @Segment.with_input_slots(data=8, length=5).with_input_checks(validate_length)
            def send_dynamic(engine: EngineProto, data: Union[Renderable, int], length: Union[Renderable, int] = 16) -> None:
                data = to_renderable(data)

                end_state = engine.create_state()

                # Unrolled loop
                for i in range(32):
                    engine.set(output, data[i])
                    if i < 31:
                        engine.jump_if(length == (i + 1), end_state, engine.next_state)  # type: ignore[arg-type]
                        engine.current_state = engine.next_state
                    else:
                        engine.jump_if(Const(1), end_state)
                engine.current_state = end_state
            ```

            As an alternative to defining the `validate_length` function, a single lambda function can be used:

            ```python
            # ...
            @Segment.with_input_slots(data=8, length=5).with_input_checks(length=lambda x: x < 32)
            def send_dynamic(engine: EngineProto, data: Union[Renderable, int], length: Union[Renderable, int] = 16) -> None:
                # ...
            ```
        """
        return SegmentFactory(
            input_checks=(input_validator, input_checks),
            input_slots=getattr(self, '_SegmentFactory__input_slots', {}),
            output_slots=getattr(self, '_SegmentFactory__output_slots', ()),
        )


class SegmentFactory(SegmentFactoryMixin):
    __slots__ = [
        '__input_checks',
        '__input_slots',
        '__output_slots',
    ]

    def __init__(
        self,
        input_checks: Tuple[Optional[Callable[..., bool]], Mapping[str, Callable[..., bool]]] = (None, {}),
        input_slots: Mapping[str, SlotSize] = {},
        output_slots: Sequence[SlotSize] = (),
    ) -> None:
        self.__input_checks = input_checks
        self.__input_slots = input_slots
        self.__output_slots = output_slots

    def __call__[T: Union[EngineProto, HelperProto], **P, R: Optional[Union[AnySignal, Sequence[AnySignal]]]](
        self, method: Callable[Concatenate[T, P], R]
    ) -> 'Segment[T, P, R]':
        return Segment(method, input_checks=self.__input_checks, input_slots=self.__input_slots, output_slots=self.__output_slots)


class Segment[T: Union[EngineProto, HelperProto], **P, R: Optional[Union[AnySignal, Sequence[AnySignal]]]](SegmentFactoryMixin):
    """Decorates a function or method as a segment.

    This decorator must either be applied to one of these options:

    1. The method of a noRTL engine.
    2. A function receiving the engine as it's argument.
    3. The method of a helper class having a noRTL engine as it's attribute `engine`.

    The method may receive input arguments. The segment will be re-created for every call with different arguments by default.

    It is possible to automatically copy all or selected inputs into the segment.
    This must be enabled by using a [@Segment.with_input_slots()][nortl.core.constructs.segment.Segment.with_input_slots] decorator.
    In this case, different inputs for the copied signals are ignored.

    Note that the method must not access any Python variables that change over time.

    Effectively, the body of the segment will be executed only once for each different call signature, to define the segment of states.
    Afterwards, the engine will add transitions to the start of the segment, and back from its end, while remembering the return address.

    Examples:
        Decorate a method of an custom engine:

        ```python
        from nortl import Engine, Segment

        class MyEngine(Engine):
            # ...

            @Segment
            def send_data(self) -> None:
                self.set(self.OUTPUT, self.INPUT)
                self.sync()

            def build(self) -> None:
                # The segment can be called as usual
                self.send_data()

                # Multiple invocations of the segment will re-use the same part (segment) of the state machine.
                self.send_data()
        ```

        Decorate a function:

        ```python
        from nortl import Engine, Segment

        engine = Engine()
        # ...

        @Segment
        def send_data(engine) -> None:
            engine.sync()
        ```

        Decorate a method of a helper class:

        ```python
        from nortl import Segment
        from nortl.core.protocols import EngineProto

        class Command:
            def __init__(self, engine: EngineProto) -> None:
                self.engine = engine

            @Segment
            def send_data(self) -> None:
                self.engine.set(self.OUTPUT, self.INPUT)
                self.engine.sync()
        ```
    """

    DEBUG_PRINT: ClassVar[bool] = False
    """Enable debug prints."""

    _engine_contexts: ClassVar[Dict[EngineProto, EngineContext]] = {}

    def __init__(
        self,
        method: Callable[Concatenate[T, P], R],
        input_slots: Mapping[str, SlotSize] = {},
        output_slots: Sequence[SlotSize] = (),
        input_checks: Tuple[Optional[Callable[P, bool]], Mapping[str, Callable[[object], bool]]] = (None, {}),
    ):
        self.__method = method
        self.__signature = signature(method)
        self.__bindings: Dict[T, Union[SegmentImplementation[T, P, R], BoundSegment[T, P, R]]] = {}
        self._input_slot_specs: Mapping[str, SlotSpec] = self._create_input_slot_specs(input_slots)
        self._output_slot_specs: Sequence[SlotSpec] = self._create_output_slot_specs(output_slots)
        self._input_checks: Sequence[InputCheck] = self._create_input_checks(*input_checks)

    def _create_input_slot_specs(self, input_slots: Mapping[str, SlotSize]) -> Mapping[str, SlotSpec]:
        """Create input slot specifications."""
        input_slot_specs = {}
        for name, width in input_slots.items():
            if name not in self.__signature.parameters:
                raise KeyError(f'{name} is not a valid argument name of the decorated function or method.')

            if isinstance(width, str):
                # Copy width from other input value
                if width not in self.__signature.parameters:
                    raise ValueError(
                        f'{width} is not a valid argument name of the decorated function or method.'
                        f'It is specified to determine the width for the input slot of argument {name}.'
                    )
            elif isinstance(width, bool) and not width:
                continue  # skip False slots
            elif isinstance(width, int):
                pass
            else:
                raise TypeError(f'Width specifier {width} for input slot of argument {name} is not of a valid type.')

            input_slot_specs[name] = SlotSpec(width)
        return input_slot_specs

    def _create_output_slot_specs(self, output_slots: Sequence[SlotSize]) -> Sequence[SlotSpec]:
        """Create output slot specifications."""
        output_slot_specs: List[SlotSpec] = []

        for pos, width in enumerate(output_slots):
            if isinstance(width, str):
                # Copy width from input value
                if width not in self.__signature.parameters:
                    raise ValueError(
                        f'{width} is not a valid argument name of the decorated function or method.'
                        f'It is specified to determine the width for the output slot at position {pos}.'
                    )
                elif (input_slot_spec := self._input_slot_specs.get(width, None)) is not None:
                    width = input_slot_spec.width  # Copy input slot specifier
            elif isinstance(width, bool) and not width:
                continue  # skip False slots
            elif isinstance(width, int):
                pass
            else:
                raise TypeError(f'Width specifier {width} for output slot at position {pos} is not of a valid type.')

            output_slot_specs.append(SlotSpec(width))

        return output_slot_specs

    def _create_input_checks(
        self, input_validator: Optional[Callable[P, bool]], input_checks: Mapping[str, Callable[..., bool]]
    ) -> Sequence[InputCheck]:
        """Create input checks."""
        callbacks: List[InputCheck] = []
        if input_validator is not None:
            callbacks.append(InputCheck.from_validator(input_validator))

        for name, callback in input_checks.items():
            callbacks.append(InputCheck.for_argument(name, callback))
        return callbacks

    # Public methods
    def __call__(self, engine: T, *args: P.args, **kwargs: P.kwargs) -> R:
        """Calls the decorated function as a segment."""
        __tracebackhide__ = True

        if engine not in self.__bindings:
            method = update_wrapper(partial(self.__method, engine), self.__method)
            self.__bindings[engine] = SegmentImplementation(self, engine, method)

        return self.__bindings[engine](*args, **kwargs)

    def inline(self, engine: T, *args: P.args, **kwargs: P.kwargs) -> R:
        """Calls the decorated function as an inlined function."""
        __tracebackhide__ = True
        return self.__method(engine, *args, **kwargs)

    @overload
    def __get__(self, obj: T, objtype: Optional[type] = ...) -> 'BoundSegment[T, P, R]': ...
    @overload
    def __get__(self, obj: None, objtype: None = ...) -> Self: ...

    def __get__(self, obj: Optional[T], objtype: Optional[type] = None) -> Union[Self, 'BoundSegment[T, P, R]']:
        # Non-data descriptor, this allows the decorator to be used on class methods
        if obj is None:
            return self
        else:
            if obj not in self.__bindings:
                self.__bindings[obj] = BoundSegment(self, obj, self.__method)
            return self.__bindings[obj]  # type: ignore[return-value]

    # Utilities
    @staticmethod
    def get_engine_context(engine: EngineProto) -> 'EngineContext':
        """Get the engine context for a specific engine.

        The engine context holds an overview of all segments.
        """
        if engine not in Segment._engine_contexts:
            Segment._engine_contexts[engine] = EngineContext(engine)
        return Segment._engine_contexts[engine]


class BoundSegment[T: Union[EngineProto, HelperProto], **P, R: Optional[Union[AnySignal, Sequence[AnySignal]]]]:
    """Segment bound to a method of an object."""

    def __init__(self, segment: Segment[T, P, R], obj: T, method: Callable[Concatenate[T, P], R]) -> None:
        self.__segment: Segment[T, P, R] = segment
        self.__obj: T = obj
        self.__method: Callable[P, R] = update_wrapper(partial(method, obj), method)
        self.__implementation: SegmentImplementation[T, P, R] = SegmentImplementation(segment, obj, self.__method, is_partial=True)

    # Mimic attributes of function segment
    @property
    def _input_slot_specs(self) -> Mapping[str, SlotSpec]:
        return self.__segment._input_slot_specs

    @property
    def _output_slot_specs(self) -> Sequence[SlotSpec]:
        return self.__segment._output_slot_specs

    @property
    def _input_checks(self) -> Sequence[InputCheck]:
        return self.__segment._input_checks

    # Public methods
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        """Calls the decorated method as a segment."""
        __tracebackhide__ = True
        return self.__implementation(*args, **kwargs)

    def inline(self, *args: P.args, **kwargs: P.kwargs) -> R:
        """Calls the decorated method as an inlined function."""
        __tracebackhide__ = True
        return self.__method(*args, **kwargs)


class SegmentImplementation[T: Union[EngineProto, HelperProto], **P, R: Optional[Union[AnySignal, Sequence[AnySignal]]]]:
    def __init__(self, segment: Union[Segment[T, P, R], BoundSegment[T, P, R]], engine: T, method: Callable[P, R], is_partial: bool = False) -> None:
        self.segment: Union[Segment[T, P, R], BoundSegment[T, P, R]] = segment
        self.engine: EngineProto = self._unpack_engine(engine)
        self.method: Callable[P, R] = method
        self.signature: Signature = signature(method)
        self.is_partial = is_partial

        # Get engine context
        self.engine_context = Segment.get_engine_context(self.engine)
        self._memory_zone: Optional[MemoryZoneProto] = None
        self.backup_addresses: Dict[WorkerProto, SignalProto] = {}
        self.rendered_segments: Dict[Tuple[WorkerProto, str], RenderedSegment] = {}
        self._active_key: Optional[Tuple[WorkerProto, str]] = None

    def _unpack_engine(self, obj: T) -> EngineProto:
        """Unpack engine from helper class instance object."""
        if hasattr(obj, 'engine'):
            return obj.engine
        else:
            return obj  # type: ignore

    @property
    def input_slot_specs(self) -> Mapping[str, SlotSpec]:
        return self.segment._input_slot_specs

    @property
    def output_slot_specs(self) -> Sequence[SlotSpec]:
        return self.segment._output_slot_specs

    @property
    def input_checks(self) -> Sequence[InputCheck]:
        return self.segment._input_checks

    # Call
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        __tracebackhide__ = True

        # Check inputs
        self.check_inputs(args, kwargs)

        if Segment.DEBUG_PRINT:
            self.engine.print(f'Calling {self.format_call(args, kwargs)}')

        with self.select_active_key(self.engine.current_worker, self.extract_signature(*args, **kwargs)) as key:
            try:
                # Check if segment with signature exists for the current worker and if not, render it
                if key not in self.rendered_segments:
                    self.render_segment(*args, **kwargs)

                # Then, call the segment
                return self.call_segment(*args, **kwargs)
            except Exception as e:
                e.add_note(f'This exception occured while calling segment {self.format_call(args, kwargs)} in {self.method.__module__}')
                raise

    # Selection of rendered segment
    @contextmanager
    def select_active_key(self, worker: WorkerProto, signature: str) -> Iterator[Tuple[WorkerProto, str]]:
        """Select active key, based on the worker and the call signature.

        The Segment will create individual rendered segments for each key.
        """
        __tracebackhide__ = True

        if self._active_key is not None:
            raise RuntimeError(f'Detected recursive call of Segment {self.method.__name__} in {self.method.__module__}. This is not supported.')

        self._active_key = key = worker, signature
        yield key
        self._active_key = None

    @property
    def active_key(self) -> Tuple[WorkerProto, str]:
        """Active key, based on the worker and the call signature."""
        if self._active_key is None:
            raise RuntimeError('Segment has no active rendered segment.')
        return self._active_key

    @property
    def active_rendered_segment(self) -> RenderedSegment:
        """Rendered segment for active key."""
        return self.rendered_segments[self.active_key]

    @active_rendered_segment.setter
    def active_rendered_segment(self, value: RenderedSegment) -> None:
        """Rendered segment for active key."""
        self.rendered_segments[self.active_key] = value

    # Check inputs
    def check_inputs[A: Sequence[object], B: Mapping[str, object]](self, args: A, kwargs: B) -> None:
        """Check inputs."""
        bound_arguments = self.bind_arguments(args=args, kwargs=kwargs)

        failed_checks: List[InputCheck] = []
        try:
            for check in self.input_checks:
                if not check(bound_arguments):
                    failed_checks.append(check)
        except Exception as e:
            raise ValueError(
                f'Failed calling segment {self.format_call(args, kwargs)}.\nInput check {check.description} raised an exception (see above).'
            ) from e

        if failed_checks:
            raise ValueError(
                f'Failed calling segment {self.format_call(args, kwargs)}.\nThe following input check(s) returned False:\n'
                + '\n'.join(f'{i + 1}. {check.description}' for i, check in enumerate(failed_checks))
            )

    # Render/call of segment
    def render_segment(self, *args: P.args, **kwargs: P.kwargs) -> None:
        """Render new segment into the noRTL engine."""

        # Save state from where the segment will be called, before adding the segment
        call_state = self.engine.current_state

        # Enter new memory zone of scratch manager
        # The input slots must be located inside the memory zone
        with self.memory_zone as memory_view:
            # Replace input arguments that need to be copied with "slot" signals
            input_slots, args, kwargs = self.add_input_slots(args, kwargs)

            # Create floating state, the transition(s) to it are added by _call_segment
            start_state = self.engine.current_state = self.engine.next_state

            # Call the decorated method to render it into the segment
            result = self.method(*args, **kwargs)

        # Prepare output slots and check if the result is safe
        output_slot_widths = self.prepare_output_slots(args, kwargs, result)

        # Save the rendered segment
        self.active_rendered_segment = RenderedSegment(
            start_state=start_state,
            end_state=self.engine.current_state,
            input_slots=input_slots,
            output_slot_widths=output_slot_widths,
            result=result,
            memory_view=memory_view,
        )

        # Return to the call state
        self.engine.current_state = call_state

        # Register the segment (once)
        self.engine_context.register_segment(self)

    def call_segment(self, *args: P.args, **kwargs: P.kwargs) -> R:
        """Call segment."""

        rendered_segment = self.active_rendered_segment

        # Generate new call ID and save it to return address
        # Call IDs are unique for each rendered segment, regardless of the call depth
        call_id = len(rendered_segment.calls)
        return_address = self.engine_context.get_return_address(min_calls=call_id + 1)
        self.push_return_address(return_address, call_id)

        # Copy the selected inputs into the input slots of the segment
        self.copy_inputs_in(args, kwargs)

        # Create output slots
        # The slot signals must be defined outside of the memory zone for the segment
        output_slots = self.create_output_slots(rendered_segment.output_slot_widths)

        # Add transition to segment
        self.engine.jump_if(Const(1), rendered_segment.start_state)

        # Expire all previous results
        for _, _, weak_references in rendered_segment.calls:
            for ref in weak_references:
                ref.expire()

        call_state = self.engine.current_state

        # Add new transition from end state for this call to return state (or the copy-out state)
        self.engine.current_state = rendered_segment.end_state
        self.engine.jump_if(return_address == call_id, self.engine.next_state)

        # Copy the selected results out of the segment and write-protect signals that are not copied
        # This will insert a sync if necessary
        result, weak_references = self.copy_result_out(rendered_segment.result, output_slots, return_address, call_id)  # type: ignore[arg-type]

        # Restore old return address
        self.engine.set(return_address, self.backup_address)

        # Switch to return state and save the calls start and return state (for debugging)
        return_state = self.engine.current_state
        rendered_segment.calls.append((call_state, return_state, weak_references))

        return result

    # Return address
    def push_return_address(self, return_address: SignalProto, call_id: int) -> None:
        """Push new call ID to return address while backing up the previous return address.

        If two calls are chained directly after each other, the assignments to the return address is changed.
        """
        # Find if the return address is already assigned in the current state
        # This happens, if two segment calls start directly after each other
        for assignment in self.engine.current_state.get_assignments(return_address):
            if not assignment.unconditional:
                raise RuntimeError(f'Return address {return_address.name} was used in a conditional assignment, indicating that it was tampered with')

            # If an assignment is found, the value will be replaced with the call ID.
            # The restored backup address is pushed to the own backup address instead. If it already belongs to the same segment, this is not needed.
            if assignment.value is not self.backup_address:
                self.engine.set(self.backup_address, assignment.value)
            assignment.value = to_renderable(call_id)
            return

        # If no assignment was found, save the call ID in the return address as usual
        self.engine.set(return_address, call_id)
        self.engine.set(self.backup_address, return_address)

    def get_fast_forward_return_address(self, return_address: SignalProto) -> SignalProto:
        """Get the return address signal for fast-forwarding copy-out."""
        # Find if the return address is already assigned in the current state
        # This happens, if two segment calls end directly after each other
        for assignment in self.engine.current_state.get_assignments(return_address):
            if not assignment.unconditional:
                raise RuntimeError(f'Return address {return_address.name} was used in a conditional assignment, indicating that it was tampered with')

            return assignment.value  # type: ignore[return-value]
        # If no assignment was found, use the return address directly
        return return_address

    # Input and output slots
    def add_input_slots[A: Sequence[object], B: Mapping[str, object]](self, args: A, kwargs: B) -> Tuple[Mapping[str, ScratchSignalProto], A, B]:
        """Add slot signals for input signals that need to be copied and replace them in the arguments."""
        new_arguments: Dict[str, object] = {}
        slot_signals: Dict[str, ScratchSignalProto] = {}

        bound_arguments = self.bind_arguments(args=args, kwargs=kwargs)
        for name, input_value in self.iter_bound_arguments(args=args, kwargs=kwargs):
            if (slot_spec := self.input_slot_specs.get(name, None)) is not None:
                # Determine width of input slot
                if isinstance(slot_spec.width, str):
                    # Dynamic width based on other input argument
                    width = bound_arguments.arguments[slot_spec.width]
                elif not isinstance(slot_spec.width, bool):
                    # Fixed integer width
                    width = slot_spec.width
                elif slot_spec.width and hasattr(input_value, 'operand_width') and input_value.operand_width is not None:
                    # Copy operand width
                    width = input_value.operand_width
                else:
                    raise ValueError(
                        f'Unable to determine the width of the input slot for argument {name}. The value {input_value} for this argument has no fixed operand_width.\n'
                        f'The slot has the following specification: {slot_spec}\n'
                    )
                new_arguments[name] = slot_signals[name] = self.engine.define_scratch(width)

            else:
                new_arguments[name] = input_value

        bound_arguments = self.signature.bind(**new_arguments)
        return slot_signals, bound_arguments.args[1:], bound_arguments.kwargs  # type: ignore[return-value]

    def copy_inputs_in[A: Sequence[object], B: Mapping[str, object]](self, args: A, kwargs: B) -> None:
        """Copy signals into the segment."""

        # Process the input values to check if they have been assigned in the current state
        # If possible, fast-forwards the values
        fast_forwarder = FastForwarder(self.engine)
        inputs: Dict[ScratchSignalProto, Renderable] = {}
        fast_forward_inputs: Dict[ScratchSignalProto, Renderable] = {}
        for name, input_value in self.iter_bound_arguments(args=args, kwargs=kwargs):
            if (input_slot := self.active_rendered_segment.input_slots.get(name, None)) is not None:
                # Attempt to convert value into renderable and check its width
                input_value = to_renderable(input_value)  # type: ignore[call-overload]
                if input_value.operand_width is not None and input_value.operand_width > (input_slot_width := input_slot.operand_width):
                    raise ValueError(
                        f'Width of input value {input_value} for argument {name} exceeds width of the input slot with '
                        f'{input_value.operand_width} vs. {input_slot_width} Bit.'
                    )
                inputs[input_slot] = input_value

                # Attempt to fast forward the value
                if (fast_forward_value := fast_forwarder(input_value)) is not None:
                    fast_forward_inputs[input_slot] = fast_forward_value

        # Add sync if needed
        if fast_forwarder.needs_sync:
            self.engine.sync()
            fast_forward_inputs = {}  # Discard all fast-forward inputs, if a sync is needed anyways

        # Recover memory view to re-enable access to the input slots and assign the input values
        with self.recover_memory_zone():
            for input_slot, input_value in inputs.items():
                input_slot.reclaim()
                self.engine.set(input_slot, fast_forward_inputs.get(input_slot, input_value))

    def prepare_output_slots[A: Sequence[object], B: Mapping[str, object]](self, args: A, kwargs: B, result: R) -> Sequence[Optional[int]]:
        """Prepare slot signals for output signals that need to be copied and return widths of the slots."""
        if result is None:
            return []

        # Pack results into iterable
        if hasattr(result, '__iter__'):
            results: Sequence[AnySignal] = result  # type: ignore
        else:
            results = (result,)  # type: ignore

        bound_arguments = self.bind_arguments(args=args, kwargs=kwargs)
        widths: List[Optional[int]] = []
        for pos, signal in enumerate(results):
            # Unwrap content of modifier
            if isinstance(signal, BaseModifier):
                signal = signal.content

            if not isinstance(signal, (Signal, SignalSlice, ScratchSignal)):
                raise TypeError(
                    f'Results of a method decorated with @Segment must be signals, got: {signal}\n'
                    'Returning any other object may lead to broken code and is therefore forbidden.'
                )

            if pos < len(self.output_slot_specs):
                slot_spec = self.output_slot_specs[pos]

                # Determine width of output slot
                if isinstance(slot_spec.width, str):
                    # Dynamic width based on input argument
                    width = bound_arguments.arguments[slot_spec.width]
                elif not isinstance(slot_spec.width, bool):
                    # Fixed integer width
                    width = slot_spec.width
                elif slot_spec.width and hasattr(signal, 'operand_width') and signal.operand_width is not None:
                    # Copy operand width
                    width = signal.operand_width
                else:
                    raise ValueError(
                        f'Unable to determine the width of the output slot for the output value at position {pos}. The output signal {signal.name} for this result has no fixed operand width.\n'
                        f'The slot has the following specification: {slot_spec}'
                    )
            else:
                width = None
            widths.append(width)
        return widths

    def create_output_slots(self, widths: Sequence[Optional[int]]) -> Sequence[Optional[ScratchSignalProto]]:
        """Create output slots for results."""
        output_slots: List[Optional[ScratchSignalProto]] = []
        for width in widths:
            if width is not None:
                # Create signals
                output_slots.append(self.engine.define_scratch(width))
            else:
                output_slots.append(None)
        return output_slots

    def copy_result_out(  # noqa: C901
        self, result: R, output_slots: Sequence[Optional[ScratchSignalProto]], return_address: SignalProto, call_id: int
    ) -> Tuple[R, Sequence[WeakReference[SignalProto]]]:
        """Copy any scratch signals in the results out of the segment body.

        This ensures that the scratch signals can be released like normal.
        """
        if result is None:
            self.engine.current_state = self.engine.next_state
            return result, ()

        # Pack results into iterable
        if is_iterable := hasattr(result, '__iter__'):
            results: Sequence[AnySignal] = result  # type: ignore
        else:
            results = (result,)  # type: ignore

        # Process the output values for output slots to check if they have been assigned in the current state
        # If possible, fast-forwards the values
        fast_forwarder = FastForwarder(self.engine)
        fast_forward_results: List[Optional[AnySignal]] = []
        for signal, output_slot in zip(results, output_slots):
            # Fast-forwarding is only needed for output slots, not signals that are directly returned
            if output_slot is not None:
                fast_forward_results.append(fast_forwarder(signal))
            else:
                fast_forward_results.append(None)

        # If fast-forwarding is impossible, everything needs to happen in an intermediary state
        if fast_forwarder.needs_sync:
            self.engine.current_state = self.engine.next_state
            fast_forward_results = [None] * len(fast_forward_results)  # Discard all fast-forward results, if a sync is needed anyways
        else:
            fast_forward_condition = self.get_fast_forward_return_address(return_address) == call_id

        new_results: List[object] = []
        weak_references: List[WeakReference[SignalProto]] = []
        # Recover memory view to re-enable access to the results
        with self.recover_memory_zone():
            for signal, fast_forward_signal, output_slot in zip(results, fast_forward_results, output_slots):
                if output_slot is not None:
                    if fast_forward_signal is not None:
                        # Reclaim all parts of an fast-forwarded expression and add or update selector assignment
                        fast_forwarder.reclaim(fast_forward_signal)
                        fast_forwarder.add_assignment_case(output_slot, fast_forward_signal, fast_forward_condition)
                    else:
                        # Reclaim the output signal
                        if isinstance(signal, ScratchSignal) and signal.released:
                            signal.reclaim()
                        self.engine.set(output_slot, signal)

                    new_results.append(output_slot)
                else:
                    ref: WeakReference[PermanentSignal] = WeakReference(signal)  # type: ignore[arg-type]
                    weak_references.append(ref)
                    new_results.append(ref)

        if fast_forwarder.needs_sync:
            self.engine.sync()
        else:
            self.engine.current_state = self.engine.next_state

        # Unpack results from iterable
        if is_iterable:
            return tuple(new_results), weak_references  # type: ignore[arg-type, return-value]
        else:
            return new_results[0], weak_references  # type: ignore[return-value]

    # Memory zone management
    @property
    def memory_zone(self) -> MemoryZoneProto:
        """Memory zone within scratch manager.

        The zone is re-used among all calls to this segment.
        """
        if self._memory_zone is None:
            self._memory_zone = self.engine.scratch_manager.create_zone(self.method.__name__)
        return self._memory_zone

    @contextmanager
    def recover_memory_zone(self) -> Iterator[None]:
        """Recover memory zone and view for the active rendered segment."""
        view = self.active_rendered_segment.memory_view
        zone = view.zone
        with zone.recover(view):
            yield

    # Return adress storage
    @property
    def backup_address(self) -> SignalProto:
        """Signal to backup previous return address during calls, for the current worker."""
        if self.engine.current_worker not in self.backup_addresses:
            # Create new backup signal
            # The name is unique for the worker and the memory zone ID (which is unique across segments)
            # The width is tied to the variable width of the return address (which is unique accross workers)
            name = self.engine.current_worker.create_scoped_name(f'BACKUP_ADDRESS_ZONE{self.memory_zone.id}')
            self.backup_addresses[self.engine.current_worker] = self.engine.define_local(
                name, width=self.engine_context.return_address.width, reset_value=0
            )
        return self.backup_addresses[self.engine.current_worker]

    # Signature binding
    def extract_signature(self, *args: P.args, **kwargs: P.kwargs) -> str:
        """Extract call signature and return it's representation as a string.

        The call signature is used as a key to distinguish if new calls of the segment can re-use an existing segment, or render a new one.
        """
        bound_arguments = self.bind_arguments(Ignore(), args=args, kwargs=kwargs)  # type: ignore[arg-type]

        # Redact arguments hidden behind input slots
        for name in self.input_slot_specs:
            bound_arguments.arguments[name] = Ignore()
        return repr(bound_arguments)

    def bind_arguments(self, obj: Optional[T] = None, /, args: Sequence[object] = (), kwargs: Mapping[str, object] = dict()) -> BoundArguments:
        """Bind arguments to signature."""
        bound_arguments = self.signature.bind(obj, *args, **kwargs)
        bound_arguments.apply_defaults()
        return bound_arguments

    def iter_bound_arguments(
        self, obj: Optional[T] = None, /, args: Sequence[object] = (), kwargs: Mapping[str, object] = dict()
    ) -> Iterator[Tuple[str, object]]:
        """Bind arguments to signature and iter over them."""
        yield from tuple(self.bind_arguments(obj, args=args, kwargs=kwargs).arguments.items())

    def format_call(self, args: Sequence[object] = (), kwargs: Mapping[str, object] = dict()) -> str:
        """Format call to method."""
        args_ = list(self.signature.parameters.keys())[0:1] if not self.is_partial else []
        args_ += [f'{arg}' for arg in args]
        args_ += [f'{key}={value}' for key, value in kwargs.items()]

        return f'{self.method.__name__}({", ".join(args_)})'



"""Verilog utility classes for rendering Verilog code.

This module provides classes for constructing Verilog declarations, modules, and
other structural elements with proper formatting.

Example:
    >>> from nortl.renderer.verilog_utils import VerilogModule, VerilogPortDeclaration, VerilogDeclaration
    >>>
    >>> # Create a module with ports and signals
    >>> module = VerilogModule("counter")
    >>> module.ports.append(VerilogPortDeclaration("input", "clk"))
    >>> module.ports.append(VerilogPortDeclaration("input", "reset"))
    >>> module.ports.append(VerilogPortDeclaration("output", "count", 4))
    >>> module.signals.append(VerilogDeclaration("reg", "count", 4))
    >>>
    >>> print(module.render())
    module counter (
        input clk,
        input reset,
        output [3:0] count
    );
    reg [3:0] count;
    endmodule
"""

from typing import Literal, Union, overload

from nortl.core.protocols import WorkerProto

from .abstractions import MultiHotEncodedStateRegister, OneHotEncodedStateRegister, StateRegister
from .structural import VerilogDeclaration, VerilogModule

__all__ = [
    'MultiHotEncodedStateRegister',
    'OneHotEncodedStateRegister',
    'StateRegister',
    'VerilogDeclaration',
    'VerilogModule',
]

ENCODINGS = Literal['binary', 'one-hot', 'multi-hot']


@overload
def create_state_var(worker: WorkerProto, encoding: Literal['binary'] = 'binary') -> StateRegister: ...
@overload
def create_state_var(worker: WorkerProto, encoding: Literal['one-hot'] = 'one-hot') -> OneHotEncodedStateRegister: ...
@overload
def create_state_var(worker: WorkerProto, encoding: Literal['multi-hot'] = 'multi-hot') -> MultiHotEncodedStateRegister: ...
def create_state_var(
    worker: WorkerProto, encoding: ENCODINGS = 'binary'
) -> Union[StateRegister, OneHotEncodedStateRegister, MultiHotEncodedStateRegister]:
    """Create state register."""
    if encoding == 'binary':
        return StateRegister(worker)
    if encoding == 'one-hot':
        return OneHotEncodedStateRegister(worker)
    if encoding == 'multi-hot':
        return MultiHotEncodedStateRegister(worker)
    raise ValueError(f'Unknown encoding: {encoding}')



from typing import List, Tuple

import networkx as nx

from nortl.core import CoreEngine


class NetworkXRenderer:
    """This class contains the methods to render a nortl to verilog code.

    It only represents states and their transitions without representing the conditions.
    """

    def __init__(self, engine: CoreEngine):
        self.engine = engine

        self.transitions: List[Tuple[str, str]] = []
        self.indent_level = 0

    def render(self) -> nx.Graph:  # type: ignore
        """Render the engine to a networkx graph."""
        g = nx.DiGraph()  # type: ignore
        for states in self.engine.states.values():
            for state in states:
                g.add_node(state.name)
                for _, next_state in state.transitions:
                    g.add_edge(state.name, next_state.name)

        return g



"""Test logical operations."""

import pytest

from nortl import All, Any
from nortl.core.engine import CoreEngine
from nortl.core.operations import BaseOperation
from nortl.core.signal import Signal


def test_any_all(a: Signal, b: Signal, c: Signal) -> None:
    """Test Any and All operations on 1 bit wide signals."""
    val: BaseOperation = Any(a, b, c)
    assert val.operand_width == 1
    assert val.render() == '(a || b || c)'

    val = All(a, b, c)
    assert val.operand_width == 1
    assert val.render() == '(a && b && c)'


def test_any_all_single(a: Signal) -> None:
    """Test Any/All of a single element."""

    val: BaseOperation = Any(a)
    assert val.operand_width == 1
    assert not val.is_constant
    assert val.render() == 'a'

    val = All(a)
    assert val.operand_width == 1
    assert not val.is_constant
    assert val.render() == 'a'

    val = Any('0b1')
    assert val.operand_width == 1
    assert val.is_constant
    assert val.render() == "1'h1"

    val = All('0b1')
    assert val.operand_width == 1
    assert val.is_constant
    assert val.render() == "1'h1"


def test_unpack_nested_any_all(a: Signal, b: Signal, c: Signal) -> None:
    """Nested Any and All operations are unpacked into a single operation."""
    val: BaseOperation = Any(Any(a, b), c)
    assert val.operand_width == 1
    assert val.render() == '(a || b || c)'

    val = All(All(a, b), c)
    assert val.operand_width == 1
    assert val.render() == '(a && b && c)'

    # Sequences of different types are not unpacked
    val = All(Any(a, b), c)
    assert val.operand_width == 1
    assert val.render() == '((a || b) && c)'

    val = Any(All(a, b), c)
    assert val.operand_width == 1
    assert val.render() == '((a && b) || c)'


def test_multibit(byte: Signal, engine: CoreEngine) -> None:
    """Test Any/All operations with multi-bit signals."""

    # Trying to use multi-bit wide signals is not allowed
    with pytest.raises(ValueError):
        Any(byte)

    # Comparison operations will result in a 1 bit wide signal
    val = Any(byte == 0, byte > 3)
    assert val.render() == '((byte == 0) || (byte > 3))'

    # Signals with parametric width are not allowed either
    WIDTH = engine.define_parameter('WIDTH', default_value=1)  # noqa: N806
    signal = engine.define_local('signal', width=WIDTH)
    with pytest.raises(ValueError):
        Any(signal)

    val = Any(signal > 8, signal == 0)
    assert val.render() == '((signal > 8) || (signal == 0))'


def test_generator_expression(engine: CoreEngine) -> None:
    """Any/All operations can be directly created from a generator expression."""
    signals = [engine.define_scratch(8) for _ in range(3)]

    val = All(signal > 0 for signal in signals)
    assert val.render() == '((SCRATCH_SIGNAL[7:0] > 0) && (SCRATCH_SIGNAL[15:8] > 0) && (SCRATCH_SIGNAL[23:16] > 0))'

    # If a generator expression is used as the first argument, no further arguments are allowed
    with pytest.raises(ValueError):
        val = All((signal > 0 for signal in signals), signals[0] > 0)  # type: ignore

    # If the signals have 1 bit width, they can be directly unpacked
    bits = [engine.define_scratch(1) for _ in range(3)]
    val = All(*bits)
    assert val.operand_width == 1
    assert val.render() == '(SCRATCH_SIGNAL[24] && SCRATCH_SIGNAL[25] && SCRATCH_SIGNAL[26])'

    # This works too
    val = All(*bits, bits[0])
    assert val.operand_width == 1
    assert val.render() == '(SCRATCH_SIGNAL[24] && SCRATCH_SIGNAL[25] && SCRATCH_SIGNAL[26] && SCRATCH_SIGNAL[24])'



import re
from typing import Dict, List, Set, Tuple, Union

from nortl.core.protocols import EngineProto, MemoryZoneProto, ScratchSignalProto


def sort_by_nat_split(x: str) -> Tuple[Union[int, str], ...]:
    """Sort string with mix of numbers and text, while preserving natural order of the numbers.

    This works by splitting string into elements and trying to convert everything to integers.
    """
    result: List[Union[int, str]] = []
    for element in re.split(r'([0-9]+)', x):
        if len(element) == 0:
            continue

        try:
            element = int(element)
        except ValueError:
            pass

        result.append(element)

    return tuple(result)


class ScratchpadVisualizationRenderer:
    """Visualization for the scratchpad over all states.

    The result can be seen as a table of the form
    | STATE    | Scratchpad Bits |
    | -------- | --------------- |
    | STATE_1  | A, B, C         |

    This will be generated as html that can be rendered inline in markdown.
    """

    def __init__(self, engine: EngineProto, include_modules: bool = True, clock_gating: bool = False):
        self.engine = engine
        self.scratch_maps: Dict[str, Dict[MemoryZoneProto, List[Tuple[str, int, int]]]] = {}  # Map state name to list of
        self.set_of_labels: Set[str] = set()
        self.list_of_labels: List[str] = []
        self.map_label_to_htmlcolor: Dict[str, str] = {}

    def _extract_frame_info(self, source: ScratchSignalProto) -> str:
        frame = source.creator_frames[-1]
        ret = f'{frame.filename}, {frame.lineno}: {frame.function}\n'
        return ret

    def _generate_map(self) -> None:
        for worker_states in self.engine.states.values():
            for state in worker_states:
                zone_map: Dict[MemoryZoneProto, List[Tuple[str, int, int]]] = dict()
                for zone in self.engine.scratch_manager.zones:
                    new_lst = []

                    for scratch_signal in state.active_scratch_signals:
                        if len(scratch_signal.creator_frames) == 0:
                            continue
                        if zone is not scratch_signal.zone:
                            continue
                        if isinstance(scratch_signal.index, slice):
                            bits = list(scratch_signal.index.indices(zone.width))

                            if bits[1] < bits[0]:
                                bits[0], bits[1] = bits[1], bits[0]

                            # Include last bit in range
                            bits[1] += 1

                            new_lst.append((self._extract_frame_info(scratch_signal), bits[0], bits[1] - bits[0]))
                        else:
                            new_lst.append((self._extract_frame_info(scratch_signal), scratch_signal.index, 1))

                        self.set_of_labels.add(self._extract_frame_info(scratch_signal))

                    zone_map[zone] = sorted(new_lst, key=lambda x: x[1])

                self.scratch_maps[state.name] = zone_map

    def generate_colors(self, n: int) -> List[str]:
        """Creats n different colors that should actually look different."""
        colors = []
        for i in range(n):
            hue = i / n
            s_, l_ = 0.65, 0.55

            def hue_to_rgb(p: float, q: float, t: float) -> float:
                if t < 0:
                    t += 1
                if t > 1:
                    t -= 1
                if t < 1 / 6:
                    return p + (q - p) * 6 * t
                if t < 1 / 2:
                    return q
                if t < 2 / 3:
                    return p + (q - p) * (2 / 3 - t) * 6
                return p

            q = l_ * (1 + s_) if l_ < 0.5 else l_ + s_ - l_ * s_
            p = 2 * l_ - q
            r = hue_to_rgb(p, q, hue + 1 / 3)
            g = hue_to_rgb(p, q, hue)
            b = hue_to_rgb(p, q, hue - 1 / 3)

            colors.append(f'#{int(r * 255):02x}{int(g * 255):02x}{int(b * 255):02x}')
        return colors

    def _create_metadata(self) -> None:
        self.list_of_labels = list(self.set_of_labels)
        for label, colorstr in zip(self.list_of_labels, self.generate_colors(len(self.list_of_labels))):
            self.map_label_to_htmlcolor[label] = colorstr

    def render(self, show_frameinfo: bool = False) -> str:
        """Will create the html table in a raw string."""
        self._generate_map()
        self._create_metadata()

        retlst = ['<table>']

        retlst.append('<tr>')
        retlst.append('<td></td>')
        retlst.extend([f'<td colspan="{zone.width}">Zone {zone.id}<br>{zone.width} Bit</td>' for zone in self.engine.scratch_manager.zones])
        retlst.append('</tr>')

        for statename in sorted(list(self.scratch_maps.keys()), key=lambda x: sort_by_nat_split(x)):
            retlst.append('<tr>')
            retlst.append(f'<td>{statename}</td>')

            for zone, entry in self.scratch_maps[statename].items():
                next_pos = 0
                for item, pos, length in entry:
                    if next_pos != pos:
                        retlst.append(f'<td colspan="{pos - next_pos}" style="background-color: #F0F0F0"></td>')
                    next_pos = pos + length

                    frameinfo = ''
                    if show_frameinfo:
                        frameinfo = item

                    if item != '':
                        retlst.append(f'<td style="background-color: {self.map_label_to_htmlcolor[item]}" colspan="{length}">{frameinfo}</td>')
                    else:
                        retlst.append('<td></td>')

                if next_pos != zone.width:
                    retlst.append(f'<td colspan="{zone.width - next_pos}" style="background-color: #F0F0F0"></td>')

            retlst.append('</tr>')

        retlst += ['</table>']
        return '\n'.join(retlst)



"""Comparison operations."""

from typing import Optional

from nortl.core.protocols import Operand
from nortl.core.renderers.operations.base import TwoSideRenderer, const_unpack

__all__ = [
    'Equality',
    'Greater',
    'GreaterOrEqual',
    'Less',
    'LessOrEqual',
    'Unequality',
]


class Equality(TwoSideRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'({self.left} == {self.right})'

    @staticmethod
    def eval(left: Operand, right: Operand) -> Optional[int]:
        """Evaluate operation into constant value."""
        if (left_ := const_unpack(left)) is not None and (right_ := const_unpack(right)) is not None:
            return left_ == right_
        return None


class Unequality(TwoSideRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'({self.left} != {self.right})'

    @staticmethod
    def eval(left: Operand, right: Operand) -> Optional[int]:
        """Evaluate operation into constant value."""
        if (left_ := const_unpack(left)) is not None and (right_ := const_unpack(right)) is not None:
            return left_ != right_
        return None


class Less(TwoSideRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'({self.left} < {self.right})'

    @staticmethod
    def eval(left: Operand, right: Operand) -> Optional[int]:
        """Evaluate operation into constant value."""
        if (left_ := const_unpack(left)) is not None and (right_ := const_unpack(right)) is not None:
            return left_ < right_
        return None


class LessOrEqual(TwoSideRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'({self.left} <= {self.right})'

    @staticmethod
    def eval(left: Operand, right: Operand) -> Optional[int]:
        """Evaluate operation into constant value."""
        if (left_ := const_unpack(left)) is not None and (right_ := const_unpack(right)) is not None:
            return left_ <= right_
        return None


class Greater(TwoSideRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'({self.left} > {self.right})'

    @staticmethod
    def eval(left: Operand, right: Operand) -> Optional[int]:
        """Evaluate operation into constant value."""
        if (left_ := const_unpack(left)) is not None and (right_ := const_unpack(right)) is not None:
            return left_ > right_
        return None


class GreaterOrEqual(TwoSideRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'({self.left} >= {self.right})'

    @staticmethod
    def eval(left: Operand, right: Operand) -> Optional[int]:
        """Evaluate operation into constant value."""
        if (left_ := const_unpack(left)) is not None and (right_ := const_unpack(right)) is not None:
            return left_ >= right_
        return None



"""Renderers.

This package contains the rendering implementation for all supported operations.

The renderers also implement the behavior for constant folding for the operations.
"""

from .arithmetics import Addition, Division, Modulo, Multiplication, Substraction
from .base import SingleRenderer, SliceRenderer, TwoSideRenderer, const_unpack
from .comparison import Equality, Greater, GreaterOrEqual, Less, LessOrEqual, Unequality
from .logic import And, ExclusiveOr, LeftShift, Or, RightShift
from .misc import Inversion, Negative, Positive
from .sequences import All, Any, Concat
from .slice import Slice
from .ternary import IfThenElse

__all__ = [
    'Addition',
    'All',
    'And',
    'Any',
    'Concat',
    'Const',
    'Division',
    'Equality',
    'ExclusiveOr',
    'Greater',
    'GreaterOrEqual',
    'IfThenElse',
    'Inversion',
    'LeftShift',
    'Less',
    'LessOrEqual',
    'Modulo',
    'Multiplication',
    'Negative',
    'Or',
    'Positive',
    'RightShift',
    'SingleRenderer',
    'Slice',
    'SliceRenderer',
    'Substraction',
    'TwoSideRenderer',
    'Unequality',
    'const_unpack',
]



# type: ignore
import pytest

from nortl.core.engine import CoreEngine
from nortl.core.operations import Var
from nortl.core.signal import Signal


@pytest.fixture()
def engine() -> CoreEngine:
    """NoRTL Engine."""
    return CoreEngine('my_engine')


@pytest.fixture
def a(engine: CoreEngine) -> Signal:
    """Signal a."""
    return Signal(engine, 'input', 'a')


@pytest.fixture
def b(engine: CoreEngine) -> Signal:
    """Signal b."""
    return Signal(engine, 'input', 'b')


@pytest.fixture
def c(engine: CoreEngine) -> Signal:
    """Signal c."""
    return Signal(engine, 'input', 'c')


@pytest.fixture
def byte(engine: CoreEngine) -> Signal:
    """8-Bit Signal byte."""
    return Signal(engine, 'input', 'byte', width=8)


@pytest.fixture
def scratch_pad(engine: CoreEngine) -> Signal:
    """Scratch pad signal."""
    return Signal(engine, 'input', 'byte', width=Var(32))



"""Utilities for constructs."""

from typing import Iterator, Optional

from nortl.core.exceptions import read_access
from nortl.core.modifiers import BaseModifier
from nortl.core.operations import BaseOperation
from nortl.core.protocols import AnySignal, AssignmentTarget, EngineProto, Renderable
from nortl.core.signal import ScratchSignal, Signal, SignalSlice
from nortl.core.state import SelectorAssignment

__all__ = [
    'FastForwarder',
]


class FastForwarder:
    """Helper class to check if values for an assignment can be fast-forwarded.

    This can be used by constructs, to avoid protective sync() cycles when passing inputs.
    """

    def __init__(self, engine: EngineProto):
        self.engine = engine
        self.needs_sync = False

    def __call__(self, value: Renderable) -> Optional[AnySignal]:
        """Check if the value for an assignment can be fast-forwarded.

        This checks if the value was freshly assigned in the current state. If this is the, case, it will return the new value.
        It will also update the attribute `needs_sync`.

        In the following example, the value of `a` is freshly assigned in the current state. It also shall be copied into a different variable, `c`, which would require a `sync()` in between:

        ```python
        engine.set(a, b)
        engine.sync()  # this protective sync() could be removed
        engine.set(c, a)
        ```

        These assignments could be transformed into:

        ```python
        engine.set(a, b)
        engine.set(c, b) # fast-forward assignment
        ```
        """
        # Check if the input value has been assigned in the current state
        fast_forward_value: AnySignal = value  # type: ignore[assignment]
        for signal in self.extract_all_signals(value):
            for assignment in self.engine.current_state.get_assignments(signal):
                overlap = signal.overlaps_with(assignment.signal)
                if overlap is False:
                    # No overlap
                    continue
                elif overlap == 'partial' or not assignment.unconditional:
                    # Partial or conditional assignments are too complicated to fast-forward
                    self.needs_sync = True
                    return None

                # If an assignment is found, the behavior depends on the type of the input value
                if isinstance(value, BaseModifier):
                    value = value.content
                if isinstance(value, (Signal, SignalSlice, ScratchSignal)):
                    # If the input value is a signal, its possible to copy the value directly into the input slot
                    # Any modifier is preserved
                    if isinstance(value, BaseModifier):
                        fast_forward_value = value.copy(assignment.value)
                    else:
                        fast_forward_value = assignment.value  # type: ignore[assignment]
                else:
                    # Signals used in operations are also too complicated to fast-forward
                    self.needs_sync = True
                    return None

        return fast_forward_value

    def reclaim(self, value: Renderable) -> None:
        """Reclaim all scratch signals in an renderable."""
        for signal in self.extract_all_signals(value):
            if isinstance(signal, ScratchSignal) and signal.released:
                signal.reclaim()

    def add_assignment_case(self, target: AssignmentTarget, value: AnySignal, condition: Renderable) -> None:
        """Set a fast-forward value to a target for a certain condition.

        This helper method will add a selector assignment to the target, or update an existing one, if possible.
        """
        for assignment in self.engine.current_state.get_assignments(target):
            overlap = target.overlaps_with(assignment.signal)

            if overlap is not True:
                # No overlap
                continue
            elif not isinstance(assignment, SelectorAssignment):
                raise RuntimeError(f'Found unexpected assignment to {target} when attempting to add assignment case: {assignment.signal}')

            # Add new case to selector
            new_selector = {condition: value for condition, value in assignment.selector.items()}
            new_selector[condition] = read_access(value)
            assignment.selector = new_selector
            return

        # If no assignment was found, add a new selector
        self.engine.set_when(target, {condition: value})

    @classmethod
    def extract_all_signals(cls, renderable: Renderable) -> Iterator[AnySignal]:
        """Extract all signals from a renderable."""
        # Unpack modifiers
        if isinstance(renderable, BaseModifier):
            renderable = renderable.content

        if isinstance(renderable, BaseOperation):
            for operand in renderable.operands:
                yield from cls.extract_all_signals(operand)
        elif isinstance(renderable, (Signal, SignalSlice, ScratchSignal)):
            yield renderable



"""Verilog code formatter for proper indentation and line breaks.

This module provides a simple formatter that adds proper indentation to Verilog code
and ensures appropriate line breaks for module and always blocks.

Example:
    >>> from nortl.renderer.verilog_utils.formatter import VerilogFormatter
    >>>
    >>> code = '''module test (
    ...     input clk,
    ...     input reset
    ... );
    ...
    ... always @(*) begin
    ...     if (reset) begin
    ...         count <= 0;
    ...     end
    ... end
    ... endmodule'''
    >>>
    >>> formatter = VerilogFormatter(code)
    >>> print(formatter.format())
    module test (
        input clk,
        input reset
    );

    always @(*) begin
        if (reset) begin
            count <= 0;
        end
    end

    endmodule
"""


class VerilogFormatter:
    r"""Format Verilog code with proper indentation and line breaks.

    The formatter automatically handles indentation for begin/end blocks and
    adds newlines after module and always declarations.

    Args:
        code: The Verilog code to format.

    Example:
        >>> formatter = VerilogFormatter("always @(*) begin\n    x <= 1;\nend")
        >>> print(formatter.format())
        always @(*) begin
            x <= 1;
        end
    """

    def __init__(self, code: str) -> None:
        """Initialize the formatter with Verilog code.

        Args:
            code: The Verilog code to format.
        """
        self.code_lst = code.split('\n')
        self.indent_level = 0

    def format(self) -> str:
        """Format the Verilog code.

        Returns:
            The formatted Verilog code with proper indentation and line breaks.
        """
        self.code_lst = [self._indent(line) for line in self.code_lst]
        self.code_lst = [self._newlines(line) for line in self.code_lst]
        return '\n'.join(self.code_lst)

    def _indent(self, line: str) -> str:
        """Indent a line based on begin/end keywords.

        Args:
            line: The line to indent.

        Returns:
            The indented line.
        """
        reasons_for_indent = ['begin', 'case']
        reasons_for_dedent = ['end', 'endcase']

        if any([x in line.split() for x in reasons_for_dedent]):
            self.indent_level -= 1

        ret = '   ' * self.indent_level + line

        if any([x in line.split() for x in reasons_for_indent]):
            self.indent_level += 1

        return ret

    def _newlines(self, line: str) -> str:
        """Add newline after module or always declarations.

        Args:
            line: The line to check.

        Returns:
            The line with a newline prefix if needed.
        """
        reasons_for_newline = ['module', 'always']
        if any(line.strip().startswith(reason) for reason in reasons_for_newline):
            return f'\n{line}'
        return line



"""Test concatenation."""

import pytest

from nortl import Concat, Const
from nortl.core.engine import CoreEngine
from nortl.core.signal import Signal


def test_concat_signal(byte: Signal) -> None:
    """Test concatenating signals."""

    # A signal and constant(s) with fixed width can be concatenated.
    val = Concat(Const('0b0000'), byte, Const('0b0000'))
    assert val.operand_width == 16
    assert val.render() == "{4'h0, byte, 4'h0}"

    # Trying to concatenate constants without fixed width will raise a ValueError
    with pytest.raises(ValueError):
        Concat(Const(0), byte, Const('0b0000'))

    # Signals without fixed width cannot be concatenated
    param = byte.engine.define_parameter('param')
    dyn_signal = byte.engine.define_local('dyn_signal', param)
    with pytest.raises(ValueError):
        Concat(Const('0b1010'), dyn_signal)

    # A workaround is to explicitely slice the signal
    val = Concat(Const('0b1010'), dyn_signal[7:0])
    assert val.operand_width == 12
    assert val.render() == "{4'hA, dyn_signal[7:0]}"


def test_concat_string_literal(byte: Signal) -> None:
    """Test using string literals."""

    # To make it easier to write constants with fixed length, the Concat also supports the same string literals, as the Const() class does.
    val = Concat('0b0000', byte, '0xCAFE', '0o7653')
    assert val.operand_width == 4 + 8 + 16 + 12
    assert val.render() == "{4'h0, byte, 16'hCAFE, 12'hFAB}"

    # String literals in decimal form will be parsed without a fixed width and will therefore raise a ValueError
    with pytest.raises(ValueError):
        Concat('120', byte)


def test_concat_parameter(byte: Signal) -> None:
    """Test concatenating constants and parameter."""
    DESIGN_ID0 = byte.engine.define_parameter('DESIGN_ID0', width=8)  # noqa: N806

    # Parameters with fixed width can be concatenated
    design_id = Concat('0x0815', DESIGN_ID0)
    assert design_id.operand_width == 24
    assert design_id.render() == "{16'h0815, DESIGN_ID0}"

    # Parameters without fixed widtth don't work
    N_COUNTERS = byte.engine.define_parameter('N_COUNTERS')  # noqa: N806
    with pytest.raises(ValueError):
        Concat(N_COUNTERS)


def test_concat_single(byte: Signal) -> None:
    """Test concatenation of a single element."""

    # If the concatenation contains only a single element, it is directly return
    val = Concat(byte)
    assert val.operand_width == 8
    assert not val.is_constant
    assert val.render() == 'byte'

    val = Concat('0b0000')
    assert val.operand_width == 4
    assert val.is_constant
    assert val.render() == "4'h0"


def test_generator_expression(engine: CoreEngine) -> None:
    """Concat operation can be directly created from a generator expression."""
    signals = [engine.define_scratch(8) for _ in range(3)]

    iterator = (signal for signal in signals)
    val = Concat(iterator)
    assert val.operand_width == 24
    assert val.render() == '{SCRATCH_SIGNAL[7:0], SCRATCH_SIGNAL[15:8], SCRATCH_SIGNAL[23:16]}'

    # If a generator expression is used as the first argument, no further arguments are allowed
    with pytest.raises(ValueError):
        val = Concat(iterator, signals[0])  # type: ignore

    # This works again
    val = Concat(*signals, signals[0])
    assert val.operand_width == 32
    assert val.render() == '{SCRATCH_SIGNAL[7:0], SCRATCH_SIGNAL[15:8], SCRATCH_SIGNAL[23:16], SCRATCH_SIGNAL[7:0]}'


def test_unpack_nested_concat(byte: Signal) -> None:
    """Nested concat operations are unpacked into a single operation."""
    val = Concat('0b1111', Concat(Const('0b0000'), byte, Const('0b0000')), byte)
    assert val.operand_width == 28
    assert val.render() == "{4'hF, 4'h0, byte, 4'h0, byte}"



"""Test constants."""

import pytest

from nortl import Const


def test_int_const() -> None:
    """Test creating constants from integers."""
    # Integer constants will have no width unless specified
    var = Const(0xFF)
    assert var.value == 0xFF
    assert var.width is None
    assert var.render() == '255'

    var = Const(0, 3)
    assert var.value == 0
    assert var.width == 3
    assert var.render() == "3'h0"


def test_bool_const() -> None:
    """Test creating constants from boolean values."""
    # Boolean constants will have a width of 1 and 0 or 1 as value
    var = Const(True)
    assert var.value == 1
    assert var.width == 1
    assert var.render() == "1'h1"

    var = Const(False)
    assert var.value == 0
    assert var.width == 1
    assert var.render() == "1'h0"

    # If an explicit width is provided, the value is padded
    var = Const(True, 16)
    assert var.value == 1
    assert var.width == 16
    assert var.render() == "16'h0001"


def test_string_const() -> None:
    """Test creating constants from strings."""
    # Constants can be parsed from binary, octal or hexadecimal numbers
    var = Const('0b101')
    assert var.value == 5
    assert var.width == 3
    assert var.render() == "3'h5"

    # The width of hexadecimal numbers will be a multiplier of 4
    var = Const('0x301')
    assert var.value == 0x301
    assert var.width == 12
    assert var.render() == "12'h301"

    # Width can be explicitely set
    var = Const('0x301', 13)
    assert var.value == 0x301
    assert var.width == 13
    assert var.render() == "13'h0301"

    # Width must be large enough to fit the value
    with pytest.raises(ValueError):
        Const('0x301', 10)

    # The width of octal numbers will be a mulitplier of 3
    var = Const('0o1')
    assert var.value == 1
    assert var.width == 3
    assert var.render() == "3'h1"

    # Constants can also be parsed from decimal numbers
    var = Const('123')
    assert var.value == 123
    assert var.width is None
    assert var.render() == '123'

    # The width of decimal numbers must be set explicitely
    var = Const('123', 32)
    assert var.value == 0x7B
    assert var.width == 32
    assert var.render() == "32'h0000007B"

    # Trying to create a constant from a non-integer string will fail
    with pytest.raises(ValueError):
        Const('miau')



"""Ternary operation."""

from typing import Optional, Union

from nortl.core.protocols import Operand, Renderable
from nortl.core.renderers.operations.base import TernaryRenderer, const_unpack

__all__ = [
    'IfThenElse',
]


class IfThenElse(TernaryRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        # If the expression was folded, both values will be identical
        if (result := self.true_value.render(target=target)) == self.false_value.render(target=target):
            return result
        else:
            return f'{self.condition} ? {self.true_value} : {self.false_value}'

    @staticmethod
    def eval(condition: Operand, true_value: Operand, false_value: Operand) -> Optional[Union[Renderable, int]]:
        """Evaluate operation into constant value or single operand.

        If the condition is constantly True or False, will return the corresponding branch.
        """
        condition_ = const_unpack(condition)
        true_value_ = const_unpack(true_value)
        false_value_ = const_unpack(false_value)

        if condition_ is None:
            return None

        if condition_ > 0:
            if true_value_ is not None:
                return true_value_  # Constant value
            return true_value  # Single Renderable
        else:
            if false_value_ is not None:
                return false_value_  # Constant value
            return false_value  # Single renderable



from pathlib import Path
from typing import List

from nortl.core.module import Module

VERILOG_LIBRARY_DIR = Path(__file__).parent

BUILT_IN_LIB = {
    'nortl_sync': (VERILOG_LIBRARY_DIR / 'nortl_sync.sv').resolve(),
    'nortl_edge_detector': (VERILOG_LIBRARY_DIR / 'nortl_edge_detector.sv').resolve(),
    'nortl_delay': (VERILOG_LIBRARY_DIR / 'nortl_delay.sv').resolve(),
    'nortl_count_down_timer': (VERILOG_LIBRARY_DIR / 'nortl_count_down_timer.sv').resolve(),
    'nortl_clock_gate': (VERILOG_LIBRARY_DIR / 'nortl_clock_gate.sv').resolve(),
}


def get_modules() -> List[Module]:
    module_list = []

    hdl_sync = ''
    with open(BUILT_IN_LIB['nortl_sync'], 'r') as file:
        hdl_sync = file.read()

    sync = Module('nortl_sync', hdl_sync)
    sync.add_port('IN')
    sync.add_port('OUT')
    sync.add_parameter('DATA_WIDTH', 1)
    sync.add_port('CLK_REQ')
    sync.set_clk_request('CLK_REQ')
    module_list.append(sync)

    hdl_edge_detect = ''
    with open(BUILT_IN_LIB['nortl_edge_detector'], 'r') as file:
        hdl_edge_detect = file.read()

    edge_detect = Module('nortl_edge_detector', hdl_edge_detect)
    edge_detect.add_port('SIGNAL')
    edge_detect.add_port('RISING')
    edge_detect.add_port('FALLING')
    edge_detect.add_port('CLK_REQ')
    edge_detect.set_clk_request('CLK_REQ')
    module_list.append(edge_detect)

    hdl_delay = ''
    with open(BUILT_IN_LIB['nortl_delay'], 'r') as file:
        hdl_delay = file.read()

    delay = Module('nortl_delay', hdl_delay)
    delay.add_port('IN')
    delay.add_port('OUT')
    delay.add_parameter('DATA_WIDTH', 1)
    delay.add_parameter('DELAY_STEPS', 1)
    delay.add_port('CLK_REQ')
    delay.set_clk_request('CLK_REQ')
    module_list.append(delay)

    hdl_timer = ''
    with open(BUILT_IN_LIB['nortl_count_down_timer'], 'r') as file:
        hdl_timer = file.read()

    timer = Module('nortl_count_down_timer', hdl_timer)
    timer.add_port('RELOAD')
    timer.add_port('ZERO')
    timer.add_port('DELAY')
    timer.add_parameter('DATA_WIDTH', 1)
    module_list.append(timer)

    cg = ''
    with open(BUILT_IN_LIB['nortl_clock_gate'], 'r') as file:
        cg = file.read()
    # Ports are not defined here -- should not be used directly!
    module_list.append(Module('nortl_clock_gate', cg))

    return module_list



"""Tracing for noRTL."""

import inspect
import sys
import tempfile
from contextlib import contextmanager
from inspect import FrameInfo
from pathlib import Path
from time import time
from typing import ClassVar, Iterator, List, Optional, Sequence, Tuple, Union

from .protocols import EngineProto, NamedEntityProto

# Optional import of pyinstrument
try:
    from pyinstrument.low_level.stat_profile_python import get_frame_info
    from pyinstrument.renderers import HTMLRenderer
    from pyinstrument.session import Session as PyinstrumentSession

    PROFILING_AVAILABLE = True

    def build_call_stack(frames: Sequence[FrameInfo]) -> List[str]:
        """Build call stack for pyinstrument."""
        return [get_frame_info(frame.frame) for frame in frames]

except ImportError:
    PROFILING_AVAILABLE = False
    PROFILING_INSTALL_NOTICE = (
        'Please install pyinstrument to enable support for profiling. '
        + 'noRTL defines it as an optional dependency, that can be installed with e.g. \'pip install "nortl[profiling]"\'.'
    )

    # Fallback placeholders
    class PyinstrumentSession:  # type: ignore[no-redef]
        def __init__(self, *args: object, **kwargs: object) -> None:
            raise RuntimeError(f'Unable to create PyinstrumentSession.\n{PROFILING_INSTALL_NOTICE}')

    def build_call_stack(frames: Sequence[FrameInfo]) -> List[str]:
        return [f'{frame.filename:{frame.lineno}}' for frame in frames]  # Return dummy call stack


__all__ = [
    'Tracer',
    'enable_tracing',
]


# Default boundary: nortl.core.engine.py
DEFAULT_BOUNDARY = str((Path(__file__).parent / 'engine.py'))


@contextmanager
def enable_tracing(lower_boundary: Union[Path, str] = DEFAULT_BOUNDARY) -> Iterator[None]:
    """Enable tracing and configure upper boundary.

    noRTL tracing is disabled by default. It can be enabled by accessing `engine.tracer.enabled`, before creating any states.
    This context manager allows to comfortably enable tracing for a noRTL engine, without having to change anything inside the code creating it.

    Tracing significantly slows down noRTL by a factor of 5 to 10. It should only be enabled if actually needed.

    Example:
        For example, if your noRTL engine is created inside a function `create_engine`, you would need to modify it to enable the tracer.

        Instead, simply use this context manager:

        ```python
        from nortl import enable_tracing

        from my_code import create_engine

        with enable_tracing():
            # Create the engine
            engine = create_engine()

        # Export the trace as a pyinstrument session
        engine.tracer.session.save('my_engine.pyisession')
        ```
    """
    # Turn lower boundary into string
    if isinstance(lower_boundary, Path):
        lower_boundary = str(lower_boundary)  # Format as platform default

    # Determine upper boundary by searching the stack for the invocation of enable_tracing()
    upper_boundary = None
    for frame in inspect.stack():
        for line in frame.code_context or ():
            if 'enable_tracing()' in line:
                upper_boundary = frame.filename
                break

    # Update boundaries and temporarily enable tracing
    Tracer.lower_boundary = lower_boundary
    Tracer.upper_boundary = upper_boundary
    Tracer.enabled = True
    yield
    Tracer.enabled = False


class Tracer:
    """NoRTL tracer.

    The tracer can save the current call stack (in a filtered form) into the metadata of objects.
    This helps with debugging where states are created in the noRTL engine.

    The tracer can also collect custom pyinstrument sessions, that can be exported or directly rendered as a flamegraph.
    """

    enabled: ClassVar = False
    """Enables tracing.

    Tracing significantly slows down noRTL by a factor of 5 to 10. It should only be enabled if actually needed.
    """

    lower_boundary: str = DEFAULT_BOUNDARY
    """Lower boundary for the filtered call stack.

    This frame and all frames below are excluded included in the filtered stack trace.

    By default, the call stack stops when it reaches any method in the module `nortl.core.engine`.
    """

    upper_boundary: Optional[str] = None
    """Upper boundary for the filtered call stack.

    This frame is the first frame that is included in the filtered stack trace.

    Typically the call stack should start, where the noRTL engine is created, but exclude e.g. the "overhead" from a Jupyter server.
    """

    def __init__(self, engine: EngineProto):
        """Initialize a new tracer.

        Arguments:
            engine: noRTL engine for this tracer.
        """
        self._engine = engine
        self._session: Optional['Session'] = self.create_session()

    @property
    def engine(self) -> EngineProto:
        """NoRTL Engine of this tracer."""
        return self._engine

    @property
    def session(self) -> 'Session':
        """Active profiling session."""
        if self._session is None:
            raise RuntimeError('Tracer has no active session.')
        return self._session

    def create_session(self) -> 'Session':
        """Create a profiling session and set it as active session.

        Returns:
            The new profiling session. It is also stored in tracer.session.
        """
        self._session = Session(self, self._create_trace())
        return self._session

    def add_metadata(self, target: NamedEntityProto, key: str, profile: bool = False) -> None:
        """Retrieve current stack trace and add it to the metadata of the target.

        The stack trace will be filtered, to exclude irrelevant information, based on the boundaries of the tracer.
        At the same time, this will append the stack trace to the current session, if `profile` is True.

        Arguments:
            target: Target where the stack trace will be stored.
            key: Metadata key for the stack trace.
            profile: If True, the full stack trace will be appended to the current session.
                This effectively controls, which stack traces are included in the session profile.
                E.g., for states, it is recommended to either profile where states are created, or where the are set as current state, but not both.
        """
        if self.enabled:
            full_frames = self._create_trace()
            frames = self._filter_trace(full_frames)
            target.set_metadata(key, frames)

            if self._session is not None and profile:
                self._session.append(full_frames)

    @property
    def current_trace(self) -> Sequence[FrameInfo]:
        full_frames = self._create_trace()
        frames = self._filter_trace(full_frames)
        return frames

    def format_metadata(self, target: NamedEntityProto, key: str) -> str:
        """Format the stack trace saved in the metadata of a target into a string.

        Arguments:
            target: Target where the stack trace is stored.
            key: Metadata key for the stack trace.

        Returns:
            Descriptive string of the stack trace.
        """

        # TODO nicer representation of the stack trace
        lines: List[str] = []
        lines.append(f'{target.__class__} {target.name} was created at:')

        for i, frame in enumerate(target.get_metadata(key)) or ():
            lines.append(f'[{i}] in {frame.filename}:{frame.lineno}')
            lines.extend(f'{line.lstrip().rstrip()}' for line in frame.code_context or ())
        return '\n'.join(lines)

    # Internals
    def _create_trace(self) -> Sequence[FrameInfo]:
        """Create full stack trace."""
        return tuple(reversed(inspect.stack()))

    def _filter_trace(self, full_frames: Sequence[FrameInfo]) -> Sequence[FrameInfo]:
        """Filter the frames in a full stack trace to exclude both the noRTL internals (lower boundary) and frames above an upper boundary.

        Also marks the frames outside boundaries as hidden for pyinstrument.

        Arguments:
            full_frames: Full frames of stack trace.

        Returns:
            Filtered stack trace, containing only the frames within boundaries.
        """
        frames = []
        hidden = self.upper_boundary is not None
        for frame in full_frames:
            if frame.filename.startswith(self.lower_boundary):
                hidden = True
            if self.upper_boundary is not None and frame.filename.startswith(self.upper_boundary):
                hidden = False

            if not hidden:
                frames.append(frame)
            if hidden:
                frame.frame.f_locals['__tracebackhide__'] = True

        return frames


class Session:
    """Tracing session.

    This session stores stack traces and can be exported into a pyinstrument Session.
    """

    def __init__(self, tracer: Tracer, start_call_stack: Sequence[FrameInfo]) -> None:
        """Initialize a new tracing session.

        Arguments:
            tracer: The associated tracer.
            start_call_stack: Starting call stack of the session, created from `reversed(inspect.stack())`.
        """
        self.tracer = tracer
        self.start_call_stack = build_call_stack(start_call_stack)
        self.frame_records: List[Tuple[List[str], float]] = []

    def append(self, frames: Sequence[FrameInfo]) -> None:
        """Append new stack trace to the session.

        Arguments:
            frames: Full frames of stack trace.
        """
        self.frame_records.append((build_call_stack(frames), 1.0))

    def export(self, description: Optional[str] = None) -> PyinstrumentSession:
        """Export frame records as pyinstrument Session object.

        Arguments:
            description: Additional description for the session.
        """
        if not PROFILING_AVAILABLE:
            raise RuntimeError(f'Unable to export Session.\n{PROFILING_INSTALL_NOTICE}')

        target_description = f'Trace for noRTL engine {self.tracer.engine.module_name}'
        if description is not None:
            target_description += f': {description}'

        return PyinstrumentSession(
            frame_records=self.frame_records,
            start_time=time(),
            duration=len(self.frame_records),
            min_interval=0.0,
            max_interval=1.0,
            sample_count=len(self.frame_records),
            start_call_stack=self.start_call_stack,
            target_description=target_description,
            cpu_time=0.0,
            sys_path=sys.path,
            sys_prefixes=PyinstrumentSession.current_sys_prefixes(),
        )

    def save(self, filepath: Union[Path, str], description: Optional[str] = None) -> None:
        """Save frame records as pyinstrument Session object.

        Arguments:
            filepath: The path to save to. Using the .pyisession extension is recommended.
            description: Additional description for the session.

        The session can be loaded and rendered to HTML with:

        ```
        pyinstrument --load <test.pyisession> -r html
        ```
        """
        session = self.export(description=description)
        session.save(filepath)

    def render(self, filepath: Optional[Union[Path, str]] = None, open_in_browser: bool = True, description: Optional[str] = None) -> Optional[Path]:
        """Render frame records as as pyinstrument HTML flamegraph.

        Arguments:
            filepath: The path to save to. Using the .html extension is recommended. If no path is provided, a temporary file created.
            open_in_browser: Directly open the file in browser.
            description: Additional description for the session.

        Returns:
            Filepath to the HTML file, if it wasn't directly openend in the browser.
        """
        session = self.export(description=description)
        renderer = HTMLRenderer()

        if open_in_browser:
            if filepath is not None:
                filepath = str(filepath)
            renderer.open_in_browser(session, output_filename=filepath)
            return None
        elif filepath is None:
            filepath = Path(tempfile.NamedTemporaryFile(suffix='.html', delete=False))
        else:
            filepath = Path(filepath)

        with open(filepath, 'w', encoding='utf8') as file:
            file.write(renderer.render(session))

        return filepath



"""Core Engine."""

from typing import Any, ClassVar, Dict, Mapping, Optional, Sequence, Set, Tuple, Union

from nortl.core.exceptions import OwnershipError, TransitionLockError, read_access, write_access
from nortl.core.manager import ScratchManager, SignalManager
from nortl.core.module import Module, ModuleInstance
from nortl.core.operations import Const, to_renderable
from nortl.core.parameter import Parameter
from nortl.core.process import Thread, Worker
from nortl.core.protocols import AssignmentTarget, ModuleProto, PermanentSignal, Renderable, Selector, StateProto, WorkerProto
from nortl.core.signal import ScratchSignal, Signal, SignalSlice
from nortl.core.state import State, selector_to_renderable
from nortl.core.tracing import Tracer

__all__ = [
    'CoreEngine',
]


class CoreEngine:
    """This class represents a Verilog module that realizes a noRTL engine.

    The construction idea of this class is, that a noRTL engine can be described using the following lists:

    * A set of ports, signals, parameters and submodules
    * A set of states
    * Each state has
        * a set of assignments of the form "<signal> = <statement>"
        * a set of conditions that describe the transition next states

    Since we want to also enable parallel running engines, it is necessary to organize the states in a dictionary.
    The dictionaries keys represent the thread and the state lists represent the states of the realized engine.
    By default, we start with a main thread.

    This class only realizes the features needed to create state machines. Events on signals have to be handled on the next level of abstraction.
    """

    MAIN_WORKER_NAME: ClassVar[str] = 'main'
    MAIN_THREAD_NAME: ClassVar[str] = 'main'

    def __init__(self, module_name: str, reset_state_name: str = 'IDLE'):
        """Initialize a engine.

        Arguments:
            module_name: Verilog module name.
            reset_state_name: Name of the reset state.
        """
        self.module_name = module_name
        self._reset_state_name = reset_state_name

        # Create tracer
        self._tracer = Tracer(self)

        # Parameters, signals and scratch signals
        self._parameters: Dict[str, Parameter] = dict()
        self._signal_manager = SignalManager(self)
        self._scratch_manager = ScratchManager(self)
        self._modules: Dict[str, Module] = {}
        self._module_instances: Dict[str, ModuleInstance] = {}
        self.state_metadata_template: Dict[str, Any] = {}

        # Workers
        self._workers: Dict[str, Worker] = {}

        # Create main worker
        self._main_worker = self.create_worker(self.MAIN_WORKER_NAME)
        self._current_worker = self.main_worker

        # Create main thread
        self._main_thread = self.main_worker.create_thread(self.MAIN_THREAD_NAME)

    # Tracing
    @property
    def tracer(self) -> Tracer:
        """Tracer for diagnostic information."""
        return self._tracer

    # State management
    @property
    def states(self) -> Mapping[str, Sequence[State]]:
        """All states of the engine.

        Contains a list of the states for each worker. They are stored in a dictionary and indexed with the worker name.
        """
        return {name: worker.states for name, worker in self.workers.items()}

    @property
    def state_names(self) -> Set[str]:
        """Set of the names of all states for this engine."""
        return set().union(*(worker.state_names for worker in self.workers.values()))

    def create_state(self, name: Optional[str] = None, allow_assignments: bool = True, metadata: Dict[str, Any] = {}) -> State:
        """Create a new state for the current worker.

        Arguments:
            name: Optional state name. If no name is provided, it defaults to '<current_worker>_STATE_x', where 'x' is the current number of states
                for the worker.

                If the current worker is not the main worker, the name of the state must be prefixed with the name of the current worker.
                The prefix is automatically added, if missing.
            allow_assignments: If the state allows assignments. This is used for internal purposes.
            metadata: Metadatafor the new state. If not given, the current self.state_metadata_template will be used.

        Returns:
            The created state.
        """
        if metadata == {}:
            metadata = self.state_metadata_template
        return self.current_worker.create_state(name=name, allow_assignments=allow_assignments, metadata=metadata)

    @property
    def current_state(self) -> State:
        """Current state.

        Returns:
            The current state.
        """
        return self.current_worker.current_state

    @current_state.setter
    def current_state(self, state: StateProto) -> None:
        """Current state.

        Arguments:
            state: The new current state.

        Raises:
            UnfinishedForwardDeclarationError: If the next state has been forward-declared and is different from the new current state.
        """
        self.current_worker.current_state = state

    @property
    def next_state(self) -> State:
        """Forward-declared next state.

        This simplifies the creation of new states for non-branching sections of the state graph (e.g. via sync() or wait_for()).
        When you use next_state you must set it as current_state, before you can set any other state.

        Returns:
            The forward-declared next state.
        """
        return self.current_worker.next_state

    @property
    def reset_state(self) -> State:
        """The reset state of the current worker.

        This is the initial state from which the current worker will start.

        Returns:
            The reset state.
        """
        return self.current_worker.reset_state

    # Worker management
    @property
    def workers(self) -> Mapping[str, Worker]:
        """Mapping of workers."""
        return self._workers

    def create_worker(self, name: Optional[str] = None) -> Worker:
        """Create a new worker.

        Arguments:
            name: Optional name for the worker. If no name is provided, it defaults to 'WORKER_x', where 'x' is the current number of workers.

        Returns:
            The created worker.
        """
        if name is None:
            name = f'WORKER_{len(self.workers.values())}'
        if name in self.workers:
            raise KeyError(f'Worker {name} already exists.')

        worker = Worker(self, name, reset_state_name=self._reset_state_name)
        self._workers[name] = worker
        return worker

    @property
    def current_worker(self) -> Worker:
        """Current worker."""
        return self._current_worker

    @current_worker.setter
    def current_worker(self, worker: WorkerProto) -> None:
        """Name of the current worker."""
        if worker.engine is not self:
            raise OwnershipError('Worker does not belong to this engine')
        self.current_worker.leave_foreground()
        self._current_worker = worker  # type: ignore[assignment]

    @property
    def main_worker(self) -> Worker:
        """Main worker."""
        return self._main_worker

    # Thread management
    @property
    def current_thread(self) -> Thread:
        """Current worker thread."""
        return self.current_worker.current_thread

    @property
    def main_thread(self) -> Thread:
        """Main thread."""
        return self._main_thread

    # Signal management
    @property
    def signal_manager(self) -> SignalManager:
        """Signal manager."""
        return self._signal_manager

    @property
    def scratch_manager(self) -> ScratchManager:
        """Scratch manager."""
        return self._scratch_manager

    @property
    def signals(self) -> Mapping[str, Signal]:
        """Mapping of signals."""
        return self.signal_manager.signals

    @property
    def combinationals(self) -> Sequence[Tuple[Signal, Renderable]]:
        """Sequence of combinational signal assignments."""
        return self.signal_manager.combinationals

    def define_input(
        self, name: str, width: Union[int, Parameter, Renderable] = 1, data_type: str = 'logic', is_synchronized: bool = False
    ) -> Signal:
        """Define an input signal.

        Arguments:
            name: Name of the signal.
            width: Width of the signal in bits.
            data_type: Data type of the signal.
            is_synchronized: If the signal is synchronous to the used clock domain

        Returns:
            The defined input signal.
        """
        return self.signal_manager.create_signal('input', name, width=width, data_type=data_type, is_synchronized=is_synchronized)

    def define_output(
        self,
        name: str,
        width: Union[int, Parameter, Renderable] = 1,
        reset_value: int = 0,
        data_type: str = 'logic',
        value: Union[Renderable, None] = None,
    ) -> Signal:
        """Define an output signal.

        The noRTL engine implements two variants of output ports: Registers and continuous / combinational assigns.
        Regsiters can be set in different states and combinational assigns are assigned statically.

        Arguments:
            name: Name of the signal.
            width: Width of the signal in bits.
            reset_value: Reset value of the signal (only for registers).
            data_type: Data type of the signal.
            value: Value that should be assigned continuously (only for combinational assigns).

        Returns:
            The defined output signal.
        """
        if value is None:
            # Register
            signal = self.signal_manager.create_signal('output', name, width=width, data_type=data_type, is_synchronized=True)
            self.main_worker.reset_state.add_assignment(signal, to_renderable(reset_value))
        else:
            # Combinational assign
            signal = self.signal_manager.create_signal(
                'output', name, width=width, data_type=data_type, is_synchronized=True, assignment=to_renderable(value)
            )
        return signal

    def define_local(
        self,
        name: str,
        width: Union[int, Parameter, Renderable] = 1,
        reset_value: int | None = None,
        data_type: str = 'logic',
        pulsing: bool = False,
        value: Union[Renderable, None] = None,
    ) -> Signal:
        """Define a local signal.

        Arguments:
            name: Name of the signal.
            width: Width of the signal in bits.
            reset_value (int | None): Reset value (if applicable)
            data_type: Data type of the signal.
            pulsing: If true, the signal automatically resets to zero if not set in the current state
            value: value that should be assigned continuously (only for combinational assigns)

        Returns:
            The defined local signal.
        """
        if value is None:
            signal = self.signal_manager.create_signal('local', name, width=width, data_type=data_type, is_synchronized=True, pulsing=pulsing)
            if reset_value is not None:
                self.main_worker.reset_state.add_assignment(signal, to_renderable(reset_value))
        else:
            # Combinational assign
            signal = self.signal_manager.create_signal(
                'local', name, width=width, data_type=data_type, is_synchronized=True, assignment=to_renderable(value)
            )
        return signal

    def define_scratch(self, width: int) -> ScratchSignal:
        """Define a scratch signal.

        Arguments:
            width: Width of the signal in bits.
        """
        return self.scratch_manager.create_signal(width)

    # Parameter Managment
    @property
    def parameters(self) -> Mapping[str, Parameter]:
        """Mapping of parameters."""
        return self._parameters

    def define_parameter(self, name: str, default_value: int = 0, width: Optional[int] = None) -> Parameter:
        """Defines a parameter for the engine that can be passed on to module instances.

        Note that the only supported datatype for the parameter is int!

        Arguments:
            name: Name of the parameter
            default_value: Value of the parameter. Defaults to 0.
            width: Optional width for the parameter.

        Returns:
            Parameter: Parameter object
        """
        if name in self.signals:
            raise KeyError(f'Parameter name {name} collides with existing signal name.')
        if name in self.parameters:
            raise KeyError(f'Parameter {name} already exists.')

        parameter = Parameter(self, name, default_value, width=width)
        self._parameters[name] = parameter
        return parameter

    # Setting outputs
    def set(self, signal: AssignmentTarget, level: Union[Renderable, int, bool]) -> None:
        """Set level of an output signal.

        This is non-blocking, you can set multiple signals at the same time.
        Use sync(), wait_for() or jump_if() to apply all signals and move to the next state.

        Arguments:
            signal: The signal to be set.
            level: The level to which the signal is set.

        Raises:
            TypeError: If the signal is not a noRTL signal.
            OwnershipError: If the signal does not belong to this noRTL engine.
            ConflictingAssignmentError: If the assignment overlaps with another assignment.
        """
        if not hasattr(signal, 'write_access'):
            raise TypeError(f'{signal} is not a valid assignment target.')
        if signal.name not in self.signals or signal.engine is not self:
            raise OwnershipError(f"Signal '{signal.name}' does not belong to this engine.")

        self.current_state.add_assignment(write_access(signal), read_access(to_renderable(level)))

    def set_when(self, signal: AssignmentTarget, selector: Selector, allow_short_circuit: bool = False) -> None:
        """Set level of an output signal based on a selection of conditions.

        This is non-blocking, you can set multiple signals at the same time.
        Use sync(), wait_for() or jump_if() to apply all signals and move to the next state.

        !!! warning
            Selector assignments bypass the check for conflicting assignments, if there is a partial overlap between two unconditional assignment!

            This means, if two overlapping slices of the same signals have conditional assignments, no [ConflictingAssignmentError][nortl.core.exceptions.ConflictingAssignmentError] is raised.

        !!! danger
            This is a power-user feature. Use only when experienced!

            The selector assign lead to race conditions when the assignment selectors depend on signals that are assigned in the same clock cycle
            from other workers. This cannot be checked for safety in the current code and will pass all existing checks silently.
            For safe use, only use selector assigns on static values, i.e. register values that cannot change in the current cycle (e.g. by other threads).

        Arguments:
            signal: The signal to be set.
            selector: Mapping of conditions to levels. When a condition is met, the signal is set to this level.
                If no condition is met, the signal is not updated. The conditions are processed in order, from top to bottom

                It is possible to define a default/fallback value by providing an entry named `'default'` in the selector.
                If no condition is met, the signal is set to the default level.
                Providing a `'default'` will treat the selector cases as being in prioritized order.
                This renders the expression as a `priority if` statement in Verilog, instead of a regular `if`, if this is enabled in the
                VerilogRenderer.
                Otherwise, the expression is treated as `unique if`. This means, that only one case at a time must be valid, but the cases may be
                treated in unsorted order by the synthesis tools.

                The level can also be a selector on its own (sub-selector), creating a nested nested.
                Note that the `'default'` value only applies to the selector mapping where it is defined, not to nested sub-selectors.
            allow_short_circuit: If the selector is not in prioritized order (because no `'default'` is provided), but one condition is found to be
                always True, the other conditions would be unreachable. This is a violation of the rules of `unique if`.
                If `allow_short_circuit` is True, noRTL will remove all other conditions, and only keep the always-True condition.
                If it is False , noRTL will raise an exception to avoid downstream issues in the Verilog code.

        Example:
            The following example defines a 2x2 Input AND into 2-Input OR gate.

            ```python
            engine = CoreEngine('my_engine')

            ao22 = engine.define_output('ao22', reset_value=0)
            a = engine.define_input('a')
            b = engine.define_input('b')
            c = engine.define_input('c')
            d = engine.define_input('d')

            engine.set_when(
                ao22,
                {
                    (a & b): 1,
                    (c & d): 1,
                    'default': 0,
                },
            )
            ```

            The example roughly translates to this Verilog code:

            ```verilog
            priority
            if ((a & b) == 1)
                ao22 <= 1;
            else
            if ((c & d) == 1)
                ao22 <= 1;
            else
                ao22 <= 0;
            ```

        Raises:
            TypeError: If the signal is not a noRTL signal.
            OwnershipError: If the signal does not belong to this noRTL engine.
            ConflictingAssignmentError: If the selector assignment fully overlaps with another assignment or if it partially overlaps with an unconditional assignment.
        """
        if not hasattr(signal, 'write_access'):
            raise TypeError(f'{signal} is not a valid assignment target.')
        if signal.name not in self.signals or signal.engine is not self:
            raise OwnershipError(f"Signal '{signal.name}' does not belong to this engine.")

        self.current_state.add_selector_assignment(
            write_access(signal), read_access(selector_to_renderable(selector)), allow_short_circuit=allow_short_circuit
        )

    def set_once(self, signal: AssignmentTarget, level: Union[Renderable, int, bool], reset_level: Union[Renderable, int, bool] = False) -> None:
        """Set level of an output signal for current state, reset it in next state.

        This is non-blocking, you can set multiple signals at the same time.
        Use sync(), wait_for() or jump_if() to apply all signals and move to the next state.

        Using this method will declare the next state. The current state will be restricted to a single transition to the next state.
        You cannot create multiple transitions to other states.
        If you need to do this, use the regular set() method to set signals in this state, and reset them in others.

        Arguments:
            signal: The signal to be set.
            level: The level to which the signal is set.
            reset_level: The level to which the signal is reset in the next state.

        Raises:
            TypeError: If the signal is not a noRTL signal.
            OwnershipError: If the signal does not belong to this noRTL engine.
        """
        if not hasattr(signal, 'write_access'):
            raise TypeError(f'{signal} is not a valid assignment target.')
        if signal.name not in self.signals or signal.engine is not self:
            raise OwnershipError(f"Signal '{signal.name}' does not belong to this engine.")

        self.current_state.add_assignment(write_access(signal), read_access(to_renderable(level)))
        self.next_state.add_assignment(signal, read_access(to_renderable(reset_level)))
        # Restrict transitions
        self.current_state._restrict_transition(self.next_state)

    # Creating transitions to other states
    def sync(self) -> None:
        """Synchronize outputs.

        This creates a new state and sets it as new current state.
        It also locks the transitions for the current state. You cannot add any more transitions to it.
        """
        self.current_state._add_transition(Const(True), self.next_state)
        self.current_state._lock_transitions()
        self.current_state = self.next_state

    def wait_for(self, condition: Renderable) -> None:
        """Wait until a condition is met.

        This creates a new state and sets it as new current state.
        It also locks the transitions for the current state. You cannot add any more transitions to it.

        Arguments:
            condition: The condition to wait for.

        Raises:
            ValueError: If the condition is not a Renderable instance.
            KeyError: If the condition signal does not exist in the current engine.
        """
        if not hasattr(condition, 'render'):
            raise ValueError('Condition must be a Renderable instance')
        try:
            condition.render()
        except AttributeError:
            raise KeyError(f"Condition '{condition}' does not exist in this engine.")
        if self.current_state.transitions:
            raise TransitionLockError('wait_for() cannot be used in combination with other conditional transitions.')

        self.current_state._add_transition(read_access(to_renderable(condition)), self.next_state)
        self.current_state._lock_transitions()
        self.current_state = self.next_state

    def jump_if(self, condition: Renderable, true_state: StateProto, false_state: Optional[StateProto] = None) -> None:
        """Jump to a certain state, if condition is met.

        If the condition is not met, the engine stays in this state. jump_if can be used multiple times, as long as false_state is omitted.
        The conditions will be evaluated in the order in which they are defined.

        If false_state is provided, a transition to this state will be added.
        After you have used jump_if with a false_state, you can no longer add more transistions.

        This method will not create new states and stay at the current state. You need to manually define the states provided to true_state or false_state.

        Arguments:
            condition: The condition to jump if.
            true_state: The state to jump to if the condition is met.
            false_state: The state to jump to if the condition is not met.

        Raises:
            ValueError: If the condition is not a Renderable instance.
            KeyError: If the condition signal does not exist in the current engine.
        """
        if not hasattr(condition, 'render'):
            raise ValueError('Condition must be a Renderable instance')
        try:
            condition.render()
        except AttributeError:
            raise KeyError(f"Condition '{condition}' does not exist in this engine.")

        if true_state not in self.current_worker.states:
            raise ValueError('Target state does not exist in current thread')

        if false_state is not None:
            if false_state not in self.current_worker.states:
                raise ValueError('Target state does not exist in current thread')

        self.current_state._add_transition(read_access(to_renderable(condition)), true_state)

        if false_state is not None:
            self.current_state._add_transition(to_renderable(~condition), false_state)
            self.current_state._lock_transitions()

    # Debugging Prints
    def print(self, line: str, *args: Renderable) -> None:
        """Adds a line to the print list of the current state that will be processed during simulation."""
        self.current_state.print(line, *args)

    def printf(self, fname: str, line: str, *args: Renderable) -> None:
        """Store an item that will be output to a file during simulation."""
        self.current_state.printf(fname, line, *args)

    # Module definition
    @property
    def modules(self) -> Mapping[str, Module]:
        """Mapping of available modules this engine."""
        return self._modules

    @property
    def module_instances(self) -> Mapping[str, ModuleInstance]:
        """Mapping of module instances for this engine."""
        return self._module_instances

    def define_module(self, name: str) -> Module:
        """Define a new module.

        Arguments:
            name: Name of the module.

        Returns:
            The defined module.

        Raises:
            KeyError: If a module with the same name already exists.
        """
        if name in self.modules:
            raise KeyError(f'Module {name} already exists.')
        module = Module(name)
        self._modules[name] = module
        return module

    def add_module(self, module: ModuleProto) -> None:
        """Add a module from a library.

        Arguments:
            module: The module to be added.

        Raises:
            KeyError: If a module with the same name already exists.
        """
        if module.name in self.modules:
            raise KeyError(f'Module {module.name} already exists.')
        self._modules[module.name] = module  # type: ignore[assignment]

    def create_module_instance(self, module_name: str, instance_name: str, clock_gating: bool = False) -> ModuleInstance:
        """Create a new instance of a module.

        Arguments:
            module_name: Name of the module.
            instance_name: Name of the instance.
            clock_gating: If the unit is to be included in clock gating (if clock gating is enabled)

        Returns:
            The created module instance.

        Raises:
            KeyError: If the module does not exist or if an instance with the same name already exists.
        """
        if module_name not in self.modules:
            raise KeyError(f'Module {module_name} does not exist.')
        if instance_name in self.module_instances:
            raise KeyError(f'Module instance {instance_name} already exists.')
        module = self.modules[module_name]
        instance = ModuleInstance(module, instance_name, clock_gating=clock_gating)
        self._module_instances[instance_name] = instance
        return instance

    def connect_module_port(self, instance_name: str, port_name: str, signal: PermanentSignal) -> None:
        """Connect a port of a module instance to a signal.

        Arguments:
            instance_name: Name of the instance.
            port_name: Name of the port.
            signal: The signal to be connected.

        Raises:
            KeyError: If the instance or port does not exist.
            TypeError: If the signal is not a signal or signal slice. Scratch signals cannot be connected to modules.
            ValueError: If the port does not exist in the module.
        """
        if instance_name not in self.module_instances:
            raise KeyError(f'Module instance {instance_name} does not exist.')
        if not isinstance(signal, (Signal, SignalSlice)):
            raise TypeError(f'Signal {signal} can not be connected to a module port.')
        instance = self.module_instances[instance_name]

        if not instance.module.has_port(port_name):
            raise ValueError(f'Port {port_name} does not exist in module instance {instance_name}.')

        instance.connect_port(port_name, signal)

    def override_module_parameter(self, instance_name: str, parameter_name: str, value: Union[int, Parameter, Renderable]) -> None:
        """Override a parameter of a module instance.

        Arguments:
            instance_name: Name of the instance.
            parameter_name: Name of the parameter.
            value: New value of the parameter.

        Raises:
            KeyError: If the instance or parameter does not exist.
        """
        if instance_name not in self.module_instances:
            raise KeyError(f'Module instance {instance_name} does not exist.')
        instance = self.module_instances[instance_name]
        instance.override_parameter(parameter_name, value)



import pytest

from nortl.core.engine import CoreEngine
from nortl.core.operations import Const
from nortl.renderer.mermaid_renderer import MermaidRenderer


def test_empty_engine() -> None:
    """Test that an empty engine renders correctly."""
    engine = CoreEngine('empty_engine')
    renderer = MermaidRenderer(engine)
    rendered_engine = renderer.render()
    assert rendered_engine.startswith('---\ntitle empty_engine\n---\nstateDiagram-v2\n')


def test_engine_with_states() -> None:
    """Test that an engine with states renders correctly."""
    engine = CoreEngine('engine_with_states')
    state1 = engine.create_state('state1')
    state2 = engine.create_state('state2')
    state1._add_transition(Const(True), state2)
    renderer = MermaidRenderer(engine)
    rendered_engine = renderer.render()
    assert 'state1 --> state2' in rendered_engine


def test_engine_with_multiple_transitions() -> None:
    """Test that an engine with multiple transitions renders correctly."""
    engine = CoreEngine('engine_with_multiple_transitions')
    state1 = engine.create_state('state1')
    state2 = engine.create_state('state2')
    state3 = engine.create_state('state3')
    state1._add_transition(Const(True), state2)
    state1._add_transition(Const(True), state3)
    renderer = MermaidRenderer(engine)
    rendered_engine = renderer.render()
    assert 'state1 --> state2' in rendered_engine
    assert 'state1 --> state3' in rendered_engine


def test_engine_with_no_transitions() -> None:
    """Test that an engine with no transitions renders correctly."""
    engine = CoreEngine('engine_with_no_transitions')
    state1 = engine.create_state('state1')  # noqa: F841
    state2 = engine.create_state('state2')  # noqa: F841
    renderer = MermaidRenderer(engine)
    rendered_engine = renderer.render()
    assert 'state1 --> state2' not in rendered_engine


def test_invalid_engine() -> None:
    """Test that an invalid engine raises an exception when rendered.

    The exception occurs because the engine is not properly initialized.
    """
    with pytest.raises(AttributeError):
        renderer = MermaidRenderer(None)  # type: ignore
        renderer.render()


def test_renderer_init() -> None:
    """Test that the MermaidRenderer initializes correctly."""
    engine = CoreEngine('test_engine')
    renderer = MermaidRenderer(engine)
    assert renderer.engine == engine


def test_renderer_render() -> None:
    """Test that the MermaidRenderer renders correctly."""
    engine = CoreEngine('test_engine')
    renderer = MermaidRenderer(engine)
    rendered_engine = renderer.render()
    assert isinstance(rendered_engine, str)


def test_renderer_transitions() -> None:
    """Test that the MermaidRenderer handles transitions correctly."""
    engine = CoreEngine('test_engine')
    state1 = engine.create_state('state1')
    state2 = engine.create_state('state2')
    state1._add_transition(Const(True), state2)
    renderer = MermaidRenderer(engine)
    rendered_engine = renderer.render()
    assert 'state1 --> state2' in rendered_engine



"""This module defines exception types for noRTL."""

from typing import Mapping, Union

from nortl.core.protocols import AssignmentTarget, Renderable, RenderableSelector

__all__ = [
    'AccessAfterReleaseError',
    'ConflictingAssignmentError',
    'ExclusiveReadError',
    'ExclusiveWriteError',
    'ForbiddenAssignmentError',
    'NonIdenticalRWError',
    'OwnershipError',
    'TransitionLockError',
    'TransitionRestrictionError',
    'UnfinishedForwardDeclarationError',
    'WriteViolationError',
    'read_access',
    'write_access',
]


# Access Violation Errors
class AccessViolationError(RuntimeError):
    """Base class for errors caused by invalid read or write access to signals or operation results."""


class ExclusiveReadError(AccessViolationError):
    """This error occurs when a signal is read by more than one thread.

    This error can be suppressed by disabling or ignoring the 'exclusive_read' check for a signal.

    To permanently disable the check, use `signal.acess_checker.disable_check('exclusive_read')`.

    To ignore the check during a read or write access, wrap the signal in a [Volatile][nortl.core.modifiers.Volatile] modifier with
    `Volatile(signal, 'exclusive_read')`.
    """


class ExclusiveWriteError(AccessViolationError):
    """This error occurs when a signal is written by more than one thread.

    This error can be suppressed by disabling or ignoring the 'exclusive_write' check for a signal.

    To permanently disable the check, use `signal.acess_checker.disable_check(''exclusive_write)`.

    To ignore the check during a read or write access, wrap the signal in a [Volatile][nortl.core.modifiers.Volatile] modifier with
    `Volatile(signal, 'exclusive_write')`.
    """


class NonIdenticalRWError(AccessViolationError):
    """This error occurs when a signal is read and written by different threads.

    This error can be suppressed by disabling or ignoring the 'identical_rw' check for a signal.

    To permanently disable the check, use `signal.acess_checker.disable_checkidentical_rw('')`.

    To ignore the check during a read or write access, wrap the signal in a [Volatile][nortl.core.modifiers.Volatile] modifier with
    `Volatile(signal, 'identical_rw')`.
    """


class AccessAfterReleaseError(AccessViolationError):
    """This error occurs when a scratch signal is accessed after being released."""


class WriteViolationError(AccessViolationError):
    """This error occurs when an input signal, or otherwise read-only signal is written."""


def read_access[T: Union[Renderable, RenderableSelector]](object: T) -> T:
    """Perform read access to renderable or renderable selector.

    This helper function hides the traceback caused during recursive read access.
    """
    try:
        if isinstance(object, Mapping):
            # Perform read access on selector
            for condition, level in object.items():
                if isinstance(condition, str) and condition == 'default':
                    pass
                else:
                    read_access(condition)
                read_access(level)
        else:
            object.read_access()
    except AccessViolationError as e:
        raise e.with_traceback(None)
    return object


def write_access[T: AssignmentTarget](object: T) -> T:
    """Perform write access to signal.

    This helper function hides the traceback caused during recursive write access.
    """
    try:
        object.write_access()
    except AccessViolationError as e:
        raise e.with_traceback(None)
    return object


# Transition Errors
class TransitionError(RuntimeError):
    """Base class for errors related to invalid transitions."""


class TransitionLockError(TransitionError):
    """This error occurs when attempting to add new transitions to a state that has been locked or would need to be locked but cannot."""


class TransitionRestrictionError(TransitionError):
    """This error occurs when attempting to add new transitions to a state that has been restricted to transition to a specific other state."""


# Assignment Errors
class AssignmentError(RuntimeError):
    """Base class for errors related to invalid assignments."""


class ForbiddenAssignmentError(AssignmentError):
    """This error occurs when attempting to add assignments to a state that does not allow it."""


class ConflictingAssignmentError(AssignmentError):
    """This error occurs when attempting to add multiple conflicting assignments to the same signal in one state."""


# Misc. Errors
class OwnershipError(RuntimeError):
    """This error occurs when noRTL instances, that don't belong to the same noRTL Engine are mixed up.

    This can occur when noRTL is used in an interactive context, e.g. a Jupyter notebook.
    """


class UnfinishedForwardDeclarationError(RuntimeError):
    """This error occurs when the worker has a forward-declared state, but the user attempts to switch the current state to a different state.

    This error is meant to prevent dead-ends in the state graph.
    """



"""Arithmetic operations."""

from typing import Optional, TypeVar, Union

from nortl.core.protocols import Operand, Renderable
from nortl.core.renderers.operations.base import TwoSideRenderer, const_unpack

__all__ = [
    'Addition',
    'Division',
    'Modulo',
    'Multiplication',
    'Substraction',
]

T_Left = TypeVar('T_Left', bound=Renderable)
T_Right = TypeVar('T_Right', bound=Renderable)


class Addition(TwoSideRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'({self.left} + {self.right})'

    @staticmethod
    def eval(left: Union[T_Left, int, bool], right: Union[T_Right, int, bool]) -> Optional[Union[T_Left, T_Right, int]]:
        """Evaluate operation into constant value or single operand.

        Short circuits for addition of 0.
        """
        left_ = const_unpack(left)
        right_ = const_unpack(right)

        # Short circuiting: x + 0 = x
        if left_ == 0:
            return right
        if right_ == 0:
            return left

        if left_ is not None and right_ is not None:
            return left_ + right_
        return None


class Substraction(TwoSideRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'({self.left} - {self.right})'

    @staticmethod
    def eval(left: Union[T_Left, int, bool], right: Union[T_Right, int, bool]) -> Optional[Union[T_Left, Renderable, int]]:
        """Evaluate operation into constant value or single operand.

        Short circuits for substraction of 0.
        """
        left_ = const_unpack(left)
        right_ = const_unpack(right)

        # Short circuiting: x - 0 = x
        if right_ == 0:
            return left
        if left_ == 0:
            return -right

        if left_ is not None and right_ is not None:
            return left_ - right_
        return None


class Multiplication(TwoSideRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'({self.left} * {self.right})'

    @staticmethod
    def eval(left: Union[T_Left, int, bool], right: Union[T_Right, int, bool]) -> Optional[Union[T_Left, T_Right, int]]:
        """Evaluate operation into constant value or single operand.

        Short circuits for multiplication with 0 or 1.
        """
        left_ = const_unpack(left)
        right_ = const_unpack(right)

        # Short circuiting: x * 0 = 0
        if left_ == 0 or right_ == 0:
            return 0
        # Short circuiting: x * 1 = x
        if left_ == 1:
            return right
        if right_ == 1:
            return left

        # Other constants
        if left_ is not None and right_ is not None:
            return left_ * right_
        return None


class Division(TwoSideRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'({self.left} / {self.right})'

    @staticmethod
    def eval(left: Union[T_Left, int, bool], right: Union[T_Right, int, bool]) -> Optional[Union[T_Left, T_Right, int]]:
        """Evaluate operation into constant value or single operand.

        Short circuits for division by 0 or 1.
        """
        dividend = const_unpack(left)
        divisor = const_unpack(right)

        # Short circuiting: x / 0 is invalid
        if divisor == 0:
            raise ZeroDivisionError(f'Expression {left} / {right} leads to division by zero.')
        # Short circuiting: 0 / x = 0
        if dividend == 0:
            return 0
        # Short circuiting: x / 1 = x
        if divisor == 1:
            return left

        # Other constants
        if dividend is not None and divisor is not None:
            return int(dividend / divisor)
        return None


class Modulo(TwoSideRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        return f'({self.left} % {self.right})'

    @staticmethod
    def eval(left: Operand, right: Operand) -> Optional[int]:
        """Evaluate operation into constant value."""
        if (left_ := const_unpack(left)) is not None and (right_ := const_unpack(right)) is not None:
            return left_ % right_
        return None



from nortl.core import CoreEngine
from nortl.core.constructs import Condition


def test_condition_enter_exit() -> None:
    """Test that the Condition context manager can be entered and exited without errors."""
    engine = CoreEngine('test_engine')
    input = engine.define_input('In')

    with Condition(engine, input == 1):
        engine.sync()


def test_dual_condition_enter_exit() -> None:
    """Test that the Condition context manager can be entered and exited without errors."""
    engine = CoreEngine('test_engine')
    input = engine.define_input('In')

    with Condition(engine, input == 1):
        engine.sync()

    with Condition(engine, input == 2):
        engine.sync()



import re
import shutil
from pathlib import Path
from subprocess import PIPE, run
from tempfile import TemporaryDirectory

from nortl import Engine
from nortl.components import Timer
from nortl.renderer.verilog_renderer import VerilogRenderer


class VerilogError(Exception):
    pass


class SimulatorError(Exception):
    pass


def execute_test(engine: Engine, tb_filename: str) -> str:
    """Helper function that uses iverilog to simulate the testbench with the engine.

    Args:
        engine (Engine): Engine to be tested
        tb_filename (str): Filename of the testbench
    """

    verilog_file = Path(__file__).parent / 'verilog' / tb_filename

    with TemporaryDirectory() as tempdir_str:
        tempdir = Path(tempdir_str)

        # Copy verilog testbench to tempdir
        shutil.copy(verilog_file, tempdir)

        # Render engine to verilog
        renderer = VerilogRenderer(engine)
        with open(tempdir / 'dut.sv', 'w') as file:
            file.write(renderer.render())

        # Save code as artifact
        with open((Path(__file__).parent / 'artifacts') / 'dut.sv', 'w') as file:
            file.write(renderer.render())

        # Execute icarus verilog
        command = ['iverilog', '-g2005-sv', 'dut.sv', tb_filename]
        result = run(command, cwd=tempdir, stdout=PIPE, stderr=PIPE, universal_newlines=True)

        if result.returncode != 0:
            raise SimulatorError(f'Icarus Verilog Compiler Error: \n {result.stdout} \n {result.stderr}')

        # run simulation
        result = run(['vvp', 'a.out'], cwd=tempdir, stdout=PIPE, stderr=PIPE, universal_newlines=True)

        shutil.copyfile(tempdir / 'out.vcd', (Path(__file__).parent / 'artifacts') / 'out.vcd')

        if result.returncode != 0:
            raise SimulatorError(f'Icarus Verilog Runner Error: \n {result.stdout} \n {result.stderr}')

        return result.stdout


def test_timer() -> None:
    """Only test instantiation and if it really compiles."""
    engine = Engine('my_engine')
    engine.define_input('IN')
    engine.define_output('OUT')
    _ = Timer(engine)

    execute_test(engine, 'tb_timer.sv')


def test_timer_with_parameter() -> None:
    """Test instantiation with passed-through and if it really compiles."""
    engine = Engine('my_engine')
    engine.define_input('IN')
    engine.define_output('OUT')
    width = engine.define_parameter('TIMER_WIDTH', 8)
    _ = Timer(engine, width)

    execute_test(engine, 'tb_timer.sv')


def test_timer_blocking_delay() -> None:
    """Test if the blocking delay works correctly."""

    for i in range(3, 10):
        engine = Engine('my_engine')
        _ = engine.define_input('IN')
        s_out = engine.define_output('OUT')
        timer = Timer(engine)

        engine.set(s_out, 0)
        engine.sync()
        engine.set(s_out, 1)
        timer.wait_delay(i)
        engine.set(s_out, 0)
        engine.sync()

        res = execute_test(engine, 'tb_timer.sv')
        print(res)
        res = re.findall(r'cycles =\s*\d*', res)[0]
        res_value = int(res.split('=')[1])

        assert res_value == i


def test_two_timers() -> None:
    """Test if the blocking delay works correctly."""

    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    s_out = engine.define_output('OUT')
    timer_1 = Timer(engine)
    timer_2 = Timer(engine)

    engine.set(s_out, 0)
    engine.sync()
    engine.set(s_out, 1)
    timer_1.wait_delay(10)
    engine.set(s_out, 0)
    engine.sync()
    engine.set(s_out, 1)
    timer_2.wait_delay(20)
    engine.set(s_out, 0)
    engine.sync()

    res = execute_test(engine, 'tb_timer.sv')
    print(res)
    res = re.findall(r'cycles =\s*\d*', res)[0]
    res_value = int(res.split('=')[1])

    assert res_value == 30



# ruff: noqa: N802

import re
import shutil
from pathlib import Path
from subprocess import PIPE, run
from tempfile import TemporaryDirectory
from typing import Tuple

import pytest

from nortl import Engine
from nortl.core.constructs import Condition, Fork, ForLoop
from nortl.core.constructs.loop import WhileLoop
from nortl.core.operations import Concat, Const
from nortl.core.protocols import SignalProto
from nortl.renderer.verilog_renderer import VerilogRenderer
from nortl.renderer.verilog_utils import ENCODINGS


class VerilogError(Exception):
    pass


class SimulatorError(Exception):
    pass


STATE_ENCODINGS = ['binary', 'one-hot', 'multi-hot']


def execute_test(engine: Engine, clock_gating: bool = False, encoding: ENCODINGS = 'binary') -> str:
    """Helper function that uses iverilog to simulate the testbench with the engine.

    Args:
        engine: engine to be tested
        clock_gating: If clock gating should be used in this test
        encoding: state encoding used for the test
    """

    verilog_file = Path(__file__).parent / 'verilog' / 'testbench.sv'

    with TemporaryDirectory() as tempdir_str:
        tempdir = Path(tempdir_str)

        # Copy verilog testbench to tempdir
        shutil.copy(verilog_file, tempdir)

        # Render engine to verilog
        renderer = VerilogRenderer(engine, clock_gating=clock_gating, encoding=encoding)
        with open(tempdir / 'dut.sv', 'w') as file:
            file.write(renderer.render())

        # Save code as artifact
        with open((Path(__file__).parent / 'artifacts') / 'dut.sv', 'w') as file:
            file.write(renderer.render())

        # Execute icarus verilog
        command = ['iverilog', '-g2005-sv', 'dut.sv', 'testbench.sv']
        result = run(command, cwd=tempdir, stdout=PIPE, stderr=PIPE, universal_newlines=True)

        if result.returncode != 0:
            raise SimulatorError(f'Icarus Verilog Compiler Error: \n {result.stdout} \n {result.stderr}')

        # run simulation
        result = run(['vvp', 'a.out'], cwd=tempdir, stdout=PIPE, stderr=PIPE, universal_newlines=True)

        shutil.copyfile(tempdir / 'out.vcd', (Path(__file__).parent / 'artifacts') / 'out.vcd')

        if result.returncode != 0:
            raise SimulatorError(f'Icarus Verilog Runner Error: \n {result.stdout} \n {result.stderr}')

        return result.stdout


@pytest.mark.parametrize('encoding', STATE_ENCODINGS)
def test_simulation_setup(encoding: ENCODINGS) -> None:
    """Test that the engine can be simulated by using an empty engine."""
    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    _ = engine.define_output('OUT', 8)
    _ = execute_test(engine, encoding=encoding)


@pytest.mark.parametrize('encoding', STATE_ENCODINGS)
def test_output_probe(encoding: ENCODINGS) -> None:
    """Tests, if the output values can actually be probed."""
    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8)
    engine.set(output, 0)
    engine.sync()
    engine.set(output, 55)
    engine.wait_for(output == 1)  # Stop execution

    result = execute_test(engine, encoding=encoding)

    print(result)

    assert 'OUT= 55' in result


@pytest.mark.parametrize('encoding', STATE_ENCODINGS)
def test_output_combinational(encoding: ENCODINGS) -> None:
    """Tests, if the output values can actually be probed."""
    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    output_old = engine.define_output('OUT2', 8)
    _ = engine.define_output('OUT', 8, value=output_old)
    engine.set(output_old, 0)
    engine.sync()
    engine.set(output_old, 55)
    engine.wait_for(output_old == 1)  # Stop execution

    result = execute_test(engine, encoding=encoding)

    print(result)

    assert 'OUT= 55' in result


@pytest.mark.parametrize('encoding', STATE_ENCODINGS)
def test_output_count(encoding: ENCODINGS) -> None:
    """Tests, if the output values can actually be probed."""
    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8)
    for i in range(16):
        engine.set(output, i)
        engine.sync()
    engine.wait_for(output == 1)  # Stop execution

    result = execute_test(engine, encoding=encoding)

    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(16)]

    assert res == expected


@pytest.mark.parametrize('encoding', STATE_ENCODINGS)
def test_output_count_ForLoop(encoding: ENCODINGS) -> None:
    """Tests, if the output values can actually be probed."""
    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8)

    engine.sync()

    with ForLoop(engine, 0, 16) as ptr:
        engine.set(output, ptr)

    engine.wait_for(output == 1)  # Stop execution

    result = execute_test(engine, encoding=encoding)

    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(16)]

    assert res == expected


@pytest.mark.parametrize('encoding', STATE_ENCODINGS)
def test_WhileLoop(encoding: ENCODINGS) -> None:
    engine = Engine('my_engine')
    IN = engine.define_input('IN')  # noqa: N806
    OUT = engine.define_output('OUT', 8)  # noqa: N806

    engine.sync()

    with WhileLoop(engine, OUT < 4):
        engine.wait_for(IN == 1)
        engine.set(OUT, OUT + 1)
        engine.sync()
        engine.wait_for(IN == 0)

    result = execute_test(engine, encoding=encoding)

    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(5)]

    assert res == expected


@pytest.mark.parametrize('encoding', STATE_ENCODINGS)
def test_slicing(encoding: ENCODINGS) -> None:
    """Tests, if renderables are sliced correctly."""
    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8)
    a = engine.define_local('a', 8, 6)
    b = engine.define_local('b', 8, 5)

    for i in range(16):
        engine.set(output, i)
        engine.sync()

    engine.set(output, (a + b)[3:0])
    engine.sync()

    engine.wait_for(output == 11)  # Stop execution

    result = execute_test(engine, encoding=encoding)

    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(16)] + ['OUT=11']
    print(res)

    assert res == expected


@pytest.mark.parametrize('encoding', STATE_ENCODINGS)
def test_slicing_of_concat(encoding: ENCODINGS) -> None:
    """Tests, if a Concat is sliced correctly."""
    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8)
    a = engine.define_local('a', 8, 6)
    b = engine.define_local('b', 8, 5)

    for i in range(16):
        engine.set(output, i)
        engine.sync()

    engine.set(output, Concat(a, b)[3:0])
    engine.sync()

    engine.wait_for(output == 11)  # Stop execution

    result = execute_test(engine, encoding=encoding)

    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(16)] + ['OUT=5']
    print(res)

    assert res == expected


@pytest.mark.parametrize('encoding', STATE_ENCODINGS)
def test_output_count_CG(encoding: ENCODINGS) -> None:
    """Tests, if the output values can actually be probed -- with enabled clock gating."""
    engine = Engine('my_engine')
    _ = engine.define_input('IN')

    engine.current_state.set_metadata('Clock_gating', True)

    output = engine.define_output('OUT', 8)
    for i in range(16):
        engine.set(output, i)
        engine.sync()
    engine.wait_for(output == 1)  # Stop execution

    result = execute_test(engine, clock_gating=True, encoding=encoding)

    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(16)]

    assert res == expected


@pytest.mark.parametrize('encoding', STATE_ENCODINGS)
def test_rising_edge(encoding: ENCODINGS) -> None:
    """Tests, if the rising-edge detection works."""
    engine = Engine('my_engine')
    input = engine.define_input('IN')
    output = engine.define_output('OUT', 8)

    for i in range(16):
        engine.wait_for(input.rising())
        engine.set(output, i)
    engine.wait_for(output == 1)  # Stop execution

    result = execute_test(engine, encoding=encoding)

    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(6)]

    print(res)
    assert res == expected


@pytest.mark.parametrize('encoding', STATE_ENCODINGS)
def test_falling_edge(encoding: ENCODINGS) -> None:
    """Tests, if the falling-edge detection works."""
    engine = Engine('my_engine')
    input = engine.define_input('IN')
    output = engine.define_output('OUT', 8)

    for i in range(16):
        engine.wait_for(input.falling())
        engine.set(output, i)
    engine.wait_for(output == 1)  # Stop execution

    result = execute_test(engine, encoding=encoding)

    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(5)]

    print(res)
    assert res == expected


@pytest.mark.parametrize('encoding', STATE_ENCODINGS)
def test_delay_1(encoding: ENCODINGS) -> None:
    """Tests, if the falling-edge detection concetenated with a delay works."""
    engine = Engine('my_engine')
    input = engine.define_input('IN')
    output = engine.define_output('OUT', 8)

    delayed_input = input.delayed(1)

    for i in range(16):
        engine.wait_for(delayed_input.falling())
        engine.set(output, i)
    engine.wait_for(output == 1)  # Stop execution

    result = execute_test(engine, encoding=encoding)
    print(result)

    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(5)]

    print(res)
    assert res == expected


@pytest.mark.parametrize('encoding', STATE_ENCODINGS)
def test_delay_2(encoding: ENCODINGS) -> None:
    """Tests, if the falling-edge detection concetenated with a delay works."""
    engine = Engine('my_engine')
    input = engine.define_input('IN')
    output = engine.define_output('OUT', 8)
    delayed_input = input.delayed(2)
    for i in range(16):
        engine.wait_for(delayed_input.falling())
        engine.set(output, i)
    engine.wait_for(output == 1)  # Stop execution

    result = execute_test(engine, encoding=encoding)

    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(5)]

    print(res)
    assert res == expected


@pytest.mark.parametrize('encoding', STATE_ENCODINGS)
def test_synchronized(encoding: ENCODINGS) -> None:
    """Tests, if the falling-edge detection concetenated with a delay works."""
    engine = Engine('my_engine')
    input = engine.define_input('IN')
    output = engine.define_output('OUT', 8)

    delayed_input = input.synchronized()

    for i in range(16):
        engine.wait_for(delayed_input.falling())
        engine.set(output, i)
    engine.wait_for(output == 1)  # Stop execution

    result = execute_test(engine, encoding=encoding)
    print(result)

    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(5)]

    print(res)
    assert res == expected


@pytest.mark.parametrize('encoding', STATE_ENCODINGS)
def test_scratch_pad(encoding: ENCODINGS) -> None:
    """Tests, if the output values can actually be probed."""
    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8, reset_value=0)
    engine.sync()

    with Fork(engine, 'Timeouttimer') as proc:
        test = engine.define_scratch(8)
        test.access_checker.disable_check('exclusive_read')
        test.access_checker.disable_check('identical_rw')
        engine.set(test, test + 1)

        with Condition(engine, test == 100):
            proc.finish()

        test.release()

    engine.set(output, test)
    with Condition(engine, output == 9):
        proc.cancel()

    result = execute_test(engine, encoding=encoding)

    print(result)

    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(10)]

    assert res == expected


@pytest.mark.parametrize('encoding', STATE_ENCODINGS)
def test_multiple_workers(encoding: ENCODINGS) -> None:
    """Tests manually instantiating multiple workers.

    This is normally not recommended, users should use a Fork instead.
    """

    # Main worker - controls the counter in the second worker
    def build_main(engine: Engine) -> Tuple[SignalProto, SignalProto]:
        # Internal control signals between main and secondary worker
        adr_incr = engine.define_local('adr_incr', reset_value=0)
        end_of_addresses = engine.define_local('end_of_addresses', reset_value=0)

        # Allow adr_incr and end_of_addresses to be shared between threads
        adr_incr.access_checker.disable_check('identical_rw')
        end_of_addresses.access_checker.disable_check('identical_rw')

        # Move out of reset
        engine.sync()

        # Increment address
        loop = engine.current_state
        engine.set_once(adr_incr, 1)
        engine.sync()

        # Loop (State 4)
        # Normally, the FSM should loop back to reset_state, but then the test would never complete
        engine.jump_if(end_of_addresses != 1, loop, engine.next_state)
        engine.current_state = engine.next_state

        return adr_incr, end_of_addresses

    # Second worker - implements the counter
    # In practice, this could run in the same worker.
    def build_counter(engine: Engine, adr_incr: SignalProto, end_of_addresses: SignalProto) -> SignalProto:
        # Local signals
        # They must be defined while the address_counter thread is selected, otherwise a IDENTICAL_RW violation will be caused
        address_counter = engine.define_local('address_counter', width=ADDRESS_COUNTER_WIDTH, reset_value=0)
        address_counter_nxt = engine.define_local('address_counter_nxt', width=ADDRESS_COUNTER_WIDTH, value=address_counter + 1)

        engine.wait_for(adr_incr)
        engine.set(address_counter, address_counter_nxt)
        engine.set(end_of_addresses, address_counter_nxt == N_ADDRESSES - 1)
        engine.jump_if(Const(1), engine.reset_state)

        return address_counter

    # Build the engine
    engine = Engine('my_engine')

    # Parameters
    N_ADDRESSES = engine.define_parameter('N_ADDRESSES', 50)  # noqa: N806
    ADDRESS_COUNTER_WIDTH = engine.define_parameter('ADDRESS_COUNTER_WIDTH', 6)  # noqa: N806

    # Signals are passed around between the two build functions
    adr_incr, end_of_addresses = build_main(engine)

    engine.current_worker = engine.create_worker('address_counter')
    engine.current_worker.create_thread()
    engine.current_thread.active = True  # FIXME thread should not need manual activation

    address_counter = build_counter(engine, adr_incr, end_of_addresses)

    # All test engines require IN and OUT
    _ = engine.define_input('IN')
    _ = engine.define_output('OUT', 8, value=address_counter)

    # Execute the test
    result = execute_test(engine, encoding=encoding)

    print(result)

    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(N_ADDRESSES.default_value)]

    assert res == expected


@pytest.mark.parametrize('encoding', STATE_ENCODINGS)
def test_selector_assignment(encoding: ENCODINGS) -> None:
    """Tests selector assignment."""
    engine = Engine('my_engine')

    _ = engine.define_input('IN')
    sig_out = engine.define_output('OUT', 8)
    sel = engine.define_local('sel', 2, reset_value=0)
    engine.sync()

    for x in range(4):
        engine.set(sel, x)
        engine.sync()

        # Effectively, the output just mirrors X, counting from 0 to 3
        # The selector uses two styles: one with default
        engine.set_when(
            sig_out,
            {
                ~sel[1]: {
                    ~sel[0]: 0,
                    'default': 1,
                },
                sel[1]: {
                    ~sel[0]: 2,
                    'default': 3,
                },
            },
        )
        engine.sync()

    result = execute_test(engine, encoding=encoding)

    print(result)

    res = re.findall(r'OUT=\s*\d+', result)
    res = [re.sub(r'\s*', '', r) for r in res]

    expected = [f'OUT={i}' for i in range(4)]

    assert res == expected



from nortl import Const, Engine


def test_reachability_all_reachable() -> None:
    e = Engine('my_engine')

    for _ in range(10):
        e.sync()

    e.reachability_analysis()

    for statelist in e.states.values():
        for state in statelist:
            assert state.get_metadata('reachable', True)


def test_reachability_unreachable_condition() -> None:
    e = Engine('my_engine')

    with e.condition(Const(0) == 1):
        e.sync()
        save_state = e.current_state
        e.sync()

    e.reachability_analysis()

    assert not save_state.get_metadata('reachable')


def test_reachability_in_fork() -> None:
    e = Engine('my_engine')

    with e.condition(Const(1) == 1):
        e.sync()
        with e.fork('test'):
            e.sync()
            save_state = e.current_state
            e.sync()
        e.sync()

    e.reachability_analysis()

    assert save_state.get_metadata('reachable')


def test_reachability_unreachable_fork() -> None:
    e = Engine('my_engine')

    with e.condition(Const(1) == 1):
        e.sync()
        with e.fork('test'):
            e.sync()
            save_state1 = e.current_state
            e.sync()
        e.sync()

    with e.condition(Const(1) == 0):
        e.sync()
        with e.fork('test2'):
            e.sync()
            save_state2 = e.current_state
            e.sync()
        e.sync()

    e.reachability_analysis()

    assert save_state1.get_metadata('reachable')
    assert not save_state2.get_metadata('reachable')


def test_reachability_prune_unreachable() -> None:
    e = Engine('my_engine')

    with e.condition(Const(0) == 1):
        e.sync()
        save_state = e.current_state
        e.sync()

    e.prune_unreachable_states()

    for statelist in e.states.values():
        assert save_state not in statelist

        for state in statelist:
            assert state.get_metadata('reachable', True)
